/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"

kbutton_t	in_left;
kbutton_t	in_right;
kbutton_t	in_forward;
kbutton_t	in_back;
kbutton_t	in_brake;
kbutton_t	in_turbo;
kbutton_t	in_cam1;
kbutton_t	in_cam2;
kbutton_t	in_cam3;
kbutton_t	in_camreset;
int			in_impulse;
qboolean	draw_camvalues;

/*
=======
KeyDown
=======
*/
void KeyDown (kbutton_t *b)
{
	int		k;
	char	*c;

	c = Cmd_Argv(1);
	if (c[0])
		k = MR_atoi(c);
	else
		k = -1;		// typed manually at the console for continuous down

	if (k == b->down[0] || k == b->down[1])
		return;		// repeating key

	if (!b->down[0])
		b->down[0] = k;
	else if (!b->down[1])
		b->down[1] = k;
	else
	{
		Con_Printf ("Three keys down for a button!\n");
		return;
	}

	if (b->state & 1)
		return;		// still down
	b->state |= 1 + 2;	// down + inpulse down
}

/*
=====
KeyUp
=====
*/
void KeyUp (kbutton_t *b)
{
	int		k;
	char	*c;

	c = Cmd_Argv(1);
	if (c[0])
		k = MR_atoi(c);
	else
	{ // typed manually at the console, assume for unsticking, so clear all
		b->down[0] = b->down[1] = 0;
		b->state = 4;	// inpulse up
		return;
	}

	if (b->down[0] == k)
		b->down[0] = 0;
	else if (b->down[1] == k)
		b->down[1] = 0;
	else
		return;		// key up without coresponding down (menu pass through)
	if (b->down[0] || b->down[1])
		return;		// some other key is still holding it down

	if (!(b->state & 1))
		return;		// still up (this should not happen)
	b->state &= ~1;		// now up
	b->state |= 4; 		// inpulse up
}

void IN_LeftDown	(void)	{KeyDown(&in_left);}
void IN_LeftUp		(void)	{KeyUp	(&in_left);}

void IN_RightDown	(void)	{KeyDown(&in_right);}
void IN_RightUp		(void)	{KeyUp	(&in_right);}

void IN_ForwardDown	(void)	{KeyDown(&in_forward);}
void IN_ForwardUp	(void)	{KeyUp	(&in_forward);}

void IN_BackDown	(void)	{KeyDown(&in_back);}
void IN_BackUp		(void)	{KeyUp	(&in_back);}

void IN_BrakeDown	(void)	{KeyDown(&in_brake);}
void IN_BrakeUp		(void)	{KeyUp	(&in_brake);}

void IN_TurboDown	(void)	{KeyDown(&in_turbo);}
void IN_TurboUp		(void)	{KeyUp	(&in_turbo);}

void IN_Cam1Down	(void)	{KeyDown(&in_cam1);}
void IN_Cam1Up		(void)	{KeyUp	(&in_cam1);}

void IN_Cam2Down	(void)	{KeyDown(&in_cam2);}
void IN_Cam2Up		(void)	{KeyUp	(&in_cam2);}

void IN_Cam3Down	(void)	{KeyDown(&in_cam3);}
void IN_Cam3Up		(void)	{KeyUp	(&in_cam3);}

void IN_CamResetDown(void)	{KeyDown(&in_camreset);}
void IN_CamResetUp	(void)	{KeyUp	(&in_camreset);}

void IN_Impulse		(void)	{in_impulse=MR_atoi (Cmd_Argv(1));}

//==========================================================================

/*
===========
CL_SendMove
===========
*/
void CL_SendMove ()
{
	extern	cvar_t	cam_forward;
	extern	cvar_t	cam_right;
	extern	cvar_t	cam_up;
	extern	cvar_t	cam_pitch;
	extern	cvar_t	cam_roll;
	extern	cvar_t	cam_yaw;
	extern	cvar_t	cam_mode;

	int			bits;
	sizebuf_t	buf;
	byte		data[128];
	qboolean	move = true;

	buf.maxsize		= 128;
	buf.cursize		= 0;
	buf.data		= data;
	bits			= 0;
	draw_camvalues	= false;

	if (cam_mode.ivalue > 3)
	{
		if (in_cam1.state & 3)
		{
			move = false;
			if (in_forward.state	& 3)	Cvar_SetValue ("cam_forward",	cam_forward.ivalue++);
			if (in_back.state		& 3)	Cvar_SetValue ("cam_forward",	cam_forward.ivalue--);
			if (in_left.state		& 3)	Cvar_SetValue ("cam_right",		cam_right.ivalue--);
			if (in_right.state		& 3)	Cvar_SetValue ("cam_right",		cam_right.ivalue++);
		}

		if (in_cam2.state & 3)
		{
			move = false;
			if (in_forward.state	& 3)	Cvar_SetValue ("cam_pitch",		cam_pitch.ivalue--);
			if (in_back.state		& 3)	Cvar_SetValue ("cam_pitch",		cam_pitch.ivalue++);
			if (in_left.state		& 3)	Cvar_SetValue ("cam_yaw",		cam_yaw.ivalue++);
			if (in_right.state		& 3)	Cvar_SetValue ("cam_yaw",		cam_yaw.ivalue--);

			if (cam_pitch.ivalue < 0)
				Cvar_SetValue ("cam_pitch", cam_pitch.ivalue+360);
			else if (cam_pitch.ivalue > 360)
				Cvar_SetValue ("cam_pitch", cam_pitch.ivalue-360);

			if (cam_yaw.ivalue < 0)
				Cvar_SetValue ("cam_yaw", cam_yaw.ivalue+360);
			else if (cam_yaw.ivalue > 360)
				Cvar_SetValue ("cam_yaw", cam_yaw.ivalue-360);

		}

		if (in_cam3.state & 3)
		{
			move = false;
			if (in_forward.state	& 3)	Cvar_SetValue ("cam_up",		cam_up.ivalue++);
			if (in_back.state		& 3)	Cvar_SetValue ("cam_up",		cam_up.ivalue--);
			if (in_left.state		& 3)	Cvar_SetValue ("cam_roll",		cam_roll.ivalue--);
			if (in_right.state		& 3)	Cvar_SetValue ("cam_roll",		cam_roll.ivalue++);

			if (cam_roll.ivalue < 0)
				Cvar_SetValue ("cam_roll", cam_roll.ivalue+360);
			else if (cam_roll.ivalue > 360)
				Cvar_SetValue ("cam_roll", cam_roll.ivalue-360);
		}

		if (in_camreset.state & 3)
		{
			Cvar_SetValue ("cam_forward",	0);
			Cvar_SetValue ("cam_right",		0);
			Cvar_SetValue ("cam_up",		0);
			Cvar_SetValue ("cam_pitch",		0);
			Cvar_SetValue ("cam_yaw",		0);
			Cvar_SetValue ("cam_roll",		0);
		}

		draw_camvalues = !move;
	}

	if (move)
	{
		if (in_forward.state	& 3)	bits |=	1;
		if (in_back.state		& 3)	bits |=	2;
		if (in_left.state		& 3)	bits |=	4;
		if (in_right.state		& 3)	bits |=	8;
		if (in_brake.state		& 3)	bits |=	16;
		if (in_turbo.state		& 3)	bits |=	32;
		if (in_impulse)					bits |=	64;
	}

	MSG_WriteByte (&buf, clc_move);

	MSG_WriteFloat (&buf, cl.mtime[0]);

	MSG_WriteByte (&buf, bits);
		
	if (in_impulse)
		MSG_WriteByte (&buf, in_impulse);

	in_forward.state	&= ~2;
	in_back.state		&= ~2;
	in_left.state		&= ~2;
	in_right.state		&= ~2;
	in_brake.state		&= ~2;
	in_turbo.state		&= ~2;
	in_cam1.state		&= ~2;
	in_cam2.state		&= ~2;
	in_cam3.state		&= ~2;
	in_camreset.state	&= ~2;
	in_impulse			= 0;

	if (cls.demoplayback)
		return;

	if (++cl.movemessages <= 2)
		return;
	
	if (NET_SendUnreliableMessage (cls.netcon, &buf) == -1)
	{
		Con_Printf ("CL_SendMove: lost server connection\n");
		CL_Disconnect ();
	}
}

/*
============
CL_InitInput
============
*/
void CL_InitInput (void)
{
	Cmd_AddCommand ("+left",	IN_LeftDown);
	Cmd_AddCommand ("-left",	IN_LeftUp);
	Cmd_AddCommand ("+right",	IN_RightDown);
	Cmd_AddCommand ("-right",	IN_RightUp);
	Cmd_AddCommand ("+forward",	IN_ForwardDown);
	Cmd_AddCommand ("-forward",	IN_ForwardUp);
	Cmd_AddCommand ("+back",	IN_BackDown);
	Cmd_AddCommand ("-back",	IN_BackUp);
	Cmd_AddCommand ("+brake",	IN_BrakeDown);
	Cmd_AddCommand ("-brake",	IN_BrakeUp);
	Cmd_AddCommand ("+turbo",	IN_TurboDown);
	Cmd_AddCommand ("-turbo",	IN_TurboUp);
	Cmd_AddCommand ("+cam1",	IN_Cam1Down);
	Cmd_AddCommand ("-cam1",	IN_Cam1Up);
	Cmd_AddCommand ("+cam2",	IN_Cam2Down);
	Cmd_AddCommand ("-cam2",	IN_Cam2Up);
	Cmd_AddCommand ("+cam3",	IN_Cam3Down);
	Cmd_AddCommand ("-cam3",	IN_Cam3Up);
	Cmd_AddCommand ("+camreset",IN_CamResetDown);
	Cmd_AddCommand ("-camreset",IN_CamResetUp);
	Cmd_AddCommand ("impulse",	IN_Impulse);
}
