/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"
#include	"../menu.h"

client_static_t	cls;
client_state_t	cl;

entity_t		cl_entities[MAX_EDICTS];
entity_t		*cl_visedicts_car[MAX_VISEDICTS_CAR];
entity_t		*cl_visedicts_brush[MAX_VISEDICTS_BRUSH];

cvar_t			_cl_name		= {"_cl_name",			"Unknown",	true};
cvar_t			_cl_color_red	= {"_cl_color_red",		"250",		true};
cvar_t			_cl_color_green	= {"_cl_color_green",	"0",		true};
cvar_t			_cl_color_blue	= {"_cl_color_blue",	"0",		true};
cvar_t			_cl_shownet		= {"_cl_shownet",		"0"};
cvar_t			gl_skids		= {"gl_skids",			"0",		true};
cvar_t			gl_smoke		= {"gl_smoke",			"0",		true};

int				cl_numvisedicts_car;
int				cl_numvisedicts_brush;

void	GL_SetupSpeedometer	(float speed);
void	GL_SetupSkids		(vec3_t neworg1, vec3_t oldorg1, vec3_t neworg2, vec3_t oldorg2);
void	GL_SetupSmoke		(vec3_t origin1, vec3_t origin2, vec3_t forward, entity_t *ent);
void	S_Play_EngineSound	(byte wavnum, int volume, vec3_t origin, float speed);
void	S_Play_3DSound		(byte wavnum, int volume, vec3_t origin);

extern	cvar_t	volume4;
extern	cvar_t	volume5;

/*
=============
CL_ClearState
=============
*/
void CL_ClearState (void)
{
	if (!sv.active)
		Host_ClearMemory ();

	SZ_Clear (&cls.message);

	MR_memset (&cl,				0, sizeof(cl));
	MR_memset (&cl_entities,	0, sizeof(cl_entities));
}

/*
=============
CL_Disconnect
=============
*/
void CL_Disconnect (void)
{
	if (cls.demoplayback)
		CL_StopPlayback ();
	else if (cls.state == ca_connected)
	{
		if (cls.demorecording)
			CL_Stop_f ();

		Con_DPrintf ("Sending clc_disconnect\n");
		SZ_Clear (&cls.message);
		MSG_WriteByte (&cls.message, clc_disconnect);
		NET_SendUnreliableMessage (cls.netcon, &cls.message);
		SZ_Clear (&cls.message);
		NET_Close (cls.netcon);

		cls.state = ca_disconnected;
		if (sv.active)
			Host_ShutdownServer(false);
	}

	cls.demoplayback = cls.timedemo = false;
	cls.signon = 0;
}

/*
===============
CL_Disconnect_f
===============
*/
void CL_Disconnect_f (void)
{
	CL_Disconnect ();

	if (sv.active)
		Host_ShutdownServer (false);
}

/*
======================
CL_EstablishConnection
======================
*/
void CL_EstablishConnection (char *host)
{
	if (cls.state == ca_dedicated)
		return;

	if (cls.demoplayback)
		return;

	CL_Disconnect ();

	cls.netcon = NET_Connect (host);
	if (!cls.netcon)
		Host_Error ("CL_Connect: connect failed\n");
	Con_DPrintf ("CL_EstablishConnection: connected to %s\n", host);
	
	cls.demonum = -1;			// not in the demo loop now
	cls.state = ca_connected;
	cls.signon = 0;				// need all the signon messages before playing
}

/*
==============
CL_SignonReply
==============
*/
void CL_SignonReply (void)
{
	extern	cvar_t	car;

	Con_DPrintf ("CL_SignonReply: %i\n", cls.signon);

	switch (cls.signon)
	{
	case 1:
		MSG_WriteByte	(&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "prespawn");
		break;
		
	case 2:		
		MSG_WriteByte	(&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("_carmodel %s", car.string));
		
		MSG_WriteByte	(&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("_carcolor %f %f %f", _cl_color_red.fvalue, _cl_color_green.fvalue, _cl_color_blue.fvalue));

		MSG_WriteByte	(&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, va("name %s", _cl_name.string));

		MSG_WriteByte	(&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "spawn");
		break;
		
	case 3:	
		MSG_WriteByte	(&cls.message, clc_stringcmd);
		MSG_WriteString (&cls.message, "begin");
		Cache_Report ();
		break;
		
	case 4:
		SCR_EndLoadingPlaque ();
		break;
	}
}

/*
===========
CL_NextDemo
===========
*/
void CL_NextDemo (void)
{
	if (cls.demonum == -1)
		return;

	SCR_BeginLoadingPlaque ();

	if (!cls.demos[cls.demonum][0] || cls.demonum == MAX_DEMOS)
	{
		cls.demonum = 0;
		if (!cls.demos[cls.demonum][0])
		{
			Con_Printf ("No demos listed with startdemos\n");
			cls.demonum = -1;
			return;
		}
	}

	Cbuf_InsertText (va("playdemo %s\n", cls.demos[cls.demonum]));
	cls.demonum++;
}

/*
============
CL_LerpPoint
============
*/
float CL_LerpPoint (void)
{
	float	f, frac;

	f = cl.mtime[0] - cl.mtime[1];
	
	if (!f || cls.timedemo || sv.active)
	{
		cl.time = cl.mtime[0];
		return 1;
	}
		
	if (f > 0.1)
	{
		cl.mtime[1] = cl.mtime[0] - 0.1;
		f = 0.1f;
	}
	frac = (cl.time - cl.mtime[1]) / f;

	if (frac < 0)
	{
		if (frac < -0.01)
			cl.time = cl.mtime[1];
		frac = 0;
	}
	else if (frac > 1)
	{
		if (frac > 1.01)
			cl.time = cl.mtime[0];
		frac = 1;
	}
		
	return frac;
}

/*
=================
CL_RelinkEntities
=================
*/
void CL_RelinkEntities (void)
{
	entity_t	*ent;
	trace_t		downtrace;
	vec3_t		delta, downmove, forward, right, up, vel, front_skid[2], rear_skid[2];
	float		frac, d, side_vel = 0, forward_vel = 0, time;
	int			i, j;

	frac = CL_LerpPoint ();

	cl_numvisedicts_car		= 0;
	cl_numvisedicts_brush	= 0;

	for (i=1,ent=cl_entities+1 ; i<cl.num_entities ; i++,ent++)
	{
		if ((ent->msgtime != cl.mtime[0]) || (!ent->model))
		{
			ent->model = NULL;
			continue;
		}

		if (ent->forcelink)
		{
			VectorCopy (ent->msg_origin[0], ent->origin);
			VectorCopy (ent->msg_angles[0], ent->angles);
		}
		else
		{
			for (j=0 ; j<3 ; j++)
			{
				delta[j] = ent->msg_origin[0][j] - ent->msg_origin[1][j];
				if (delta[j] > 100 || delta[j] < -100)
					frac = 1;
			}

			for (j=0 ; j<3 ; j++)
			{
				ent->origin[j] = ent->msg_origin[1][j] + frac*delta[j];

				d = ent->msg_angles[0][j] - ent->msg_angles[1][j];
				if (d > 180)
					d -= 360;
				else if (d < -180)
					d += 360;
				ent->angles[j] = ent->msg_angles[1][j] + frac*d;
			}

			if (ent->model->type == mod_car)
			{
				time = cl.mtime[0] - cl.mtime[1];

				if (time != 0)
				{
					time = 1 / time;

					AngleVectors (ent->angles, forward, right, up);

					vel[0]			= (ent->msg_origin[0][0] - ent->msg_origin[1][0]) * time;
					vel[1]			= (ent->msg_origin[0][1] - ent->msg_origin[1][1]) * time;
					vel[2]			= 0;

					rear_skid[0][0]		= ent->origin[0] - forward[0] * 6.5 - right[0] * 4 - up[0] * 3.8;
					rear_skid[0][1]		= ent->origin[1] - forward[1] * 6.5 - right[1] * 4 - up[1] * 3.8;
					rear_skid[0][2]		= ent->origin[2] - forward[2] * 6.5 - right[2] * 4 - up[2] * 3.8;

					rear_skid[1][0]		= ent->origin[0] - forward[0] * 6.5 + right[0] * 4 - up[0] * 3.8;
					rear_skid[1][1]		= ent->origin[1] - forward[1] * 6.5 + right[1] * 4 - up[1] * 3.8;
					rear_skid[1][2]		= ent->origin[2] - forward[2] * 6.5 + right[2] * 4 - up[2] * 3.8;

					front_skid[0][0]	= ent->origin[0] + forward[0] * 6.5 - right[0] * 4 - up[0] * 3.8;
					front_skid[0][1]	= ent->origin[1] + forward[1] * 6.5 - right[1] * 4 - up[1] * 3.8;
					front_skid[0][2]	= ent->origin[2] + forward[2] * 6.5 - right[2] * 4 - up[2] * 3.8;

					front_skid[1][0]	= ent->origin[0] + forward[0] * 6.5 + right[0] * 4 - up[0] * 3.8;
					front_skid[1][1]	= ent->origin[1] + forward[1] * 6.5 + right[1] * 4 - up[1] * 3.8;
					front_skid[1][2]	= ent->origin[2] + forward[2] * 6.5 + right[2] * 4 - up[2] * 3.8;

					MR_memset (&downtrace, 0, sizeof(downtrace));
					VectorCopy (ent->origin, downmove);
					downmove[2] -= 4096;

					RecursiveHullCheckInfo.hull		= cl.worldmodel->hulls;
					RecursiveHullCheckInfo.trace	= &downtrace;
					VectorCopy	   (ent->origin, RecursiveHullCheckInfo.start);
					VectorSubtract (downmove, ent->origin, RecursiveHullCheckInfo.dist);

					SV_RecursiveHullCheck (0, 0, 1, ent->origin, downmove);

					VectorCopy (downtrace.plane.normal, ent->ground_normal);
					VectorCopy (downtrace.endpos, ent->ground_pos);

					ent->inair = (ent->origin[2] - downtrace.endpos[2] > 8) ? true : false;

					if (!ent->inair)
					{
						side_vel	= DotProduct (vel, right);
						
						if (side_vel < 0)
							side_vel = -side_vel;
							
						if (side_vel > 100)
						{
							GL_SetupSkids (front_skid[0], ent->front_skid[0], front_skid[1], ent->front_skid[1]);
							GL_SetupSkids (rear_skid[0],  ent->rear_skid[0],  rear_skid[1],  ent->rear_skid[1]);
							GL_SetupSmoke (rear_skid[0], rear_skid[1], forward, ent);

							if (cl.time > ent->time_sound)
							{
								if (i == cl.viewentity)
									S_Play_3DSound (4, volume4.ivalue, ent->origin);
							}
						}
					}

					if (i == cl.viewentity)
					{
						if (cl.time > ent->time_sound)
						{
							forward_vel	= DotProduct (vel, forward);

							if (forward_vel < 0)
								forward_vel = -forward_vel;

							GL_SetupSpeedometer (forward_vel);

							forward_vel *= 0.0035f;

							S_Play_EngineSound (7, volume5.ivalue, ent->origin, forward_vel);
						}
					}

					if (cl.time > ent->time_sound || (cl.time + 0.1) < ent->time_sound)
						ent->time_sound = cl.time + 0.1;

					VectorCopy (rear_skid[0],  ent->rear_skid[0]);
					VectorCopy (rear_skid[1],  ent->rear_skid[1]);
					VectorCopy (front_skid[0], ent->front_skid[0]);
					VectorCopy (front_skid[1], ent->front_skid[1]);
				}
			}
		}

		if (ent->model->type == mod_car)
		{
			ent->forcelink = false;

			if (cl_numvisedicts_car < MAX_VISEDICTS_CAR)
			{
				cl_visedicts_car[cl_numvisedicts_car] = ent;
				cl_numvisedicts_car++;
			}
		}
		else if (ent->model->type == mod_brush)
		{
			ent->forcelink = false;

			if (cl_numvisedicts_brush < MAX_VISEDICTS_BRUSH)
			{
				cl_visedicts_brush[cl_numvisedicts_brush] = ent;
				cl_numvisedicts_brush++;
			}
		}
	}
}

/*
=================
CL_ReadFromServer
=================
*/
int CL_ReadFromServer (void)
{
	int		ret;

	cl.oldtime = cl.time;
	cl.time += host_frametime;
	
	do
	{
		ret = CL_GetMessage ();
		if (ret == -1)
			Host_Error ("CL_ReadFromServer: lost server connection");
		if (!ret)
			break;
		
		cl.last_received_message = realtime;
		CL_ParseServerMessage ();
	} while (ret && cls.state == ca_connected);
	
	if (_cl_shownet.ivalue)
		Con_Printf ("\n");

	CL_RelinkEntities ();

	return 0;
}

/*
==========
CL_SendCmd
==========
*/
void CL_SendCmd (void)
{
	if (cls.state != ca_connected)
		return;

	if (cls.signon == SIGNONS)
	{
		CL_SendMove ();
	}

	if (cls.demoplayback)
	{
		SZ_Clear (&cls.message);
		return;
	}
	
	if (!cls.message.cursize)
		return;
	
	if (!NET_CanSendMessage (cls.netcon))
	{
		Con_DPrintf ("CL_WriteToServer: can't send\n");
		return;
	}

	if (NET_SendMessage (cls.netcon, &cls.message) == -1)
		Host_Error ("CL_WriteToServer: lost server connection");

	SZ_Clear (&cls.message);
}

/*
=======
CL_Init
=======
*/
void CL_Init (void)
{	
	SZ_Alloc (&cls.message, 1024);
	CL_InitInput ();
	
	Cvar_RegisterVariable (&_cl_name);
	Cvar_RegisterVariable (&_cl_color_red);
	Cvar_RegisterVariable (&_cl_color_green);
	Cvar_RegisterVariable (&_cl_color_blue);
	Cvar_RegisterVariable (&_cl_shownet);
	Cvar_RegisterVariable (&gl_skids);
	Cvar_RegisterVariable (&gl_smoke);

	Cmd_AddCommand ("disconnect",	CL_Disconnect_f);
	Cmd_AddCommand ("record",		CL_Record_f);
	Cmd_AddCommand ("stop",			CL_Stop_f);
	Cmd_AddCommand ("playdemo",		CL_PlayDemo_f);
	Cmd_AddCommand ("timedemo",		CL_TimeDemo_f);
}
