/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __COMMON__
#define __COMMON__

#include	<stdio.h>

#if !defined BYTE_DEFINED
typedef unsigned char 		byte;
#define BYTE_DEFINED 1
#endif

#undef true
#undef false

typedef enum {false, true}	qboolean;

//============================================================================

typedef struct sizebuf_s
{
	int			maxsize;
	int			cursize;
	byte		*data;
	qboolean	allowoverflow;	// if false, do a Sys_Error
	qboolean	overflowed;		// set to true if the buffer size failed
} sizebuf_t;

void	SZ_Alloc (sizebuf_t *buf, int startsize);
void	SZ_Clear (sizebuf_t *buf);
void	SZ_Write (sizebuf_t *buf, void *data, int length);
void	SZ_Print (sizebuf_t *buf, char *data);	// strcats onto the sizebuf
void	*SZ_GetSpace (sizebuf_t *buf, int length);

//============================================================================

typedef struct link_s
{
	struct link_s	*prev, *next;
} link_t;


void ClearLink (link_t *l);
void RemoveLink (link_t *l);
void InsertLink (link_t *l, link_t *before);

#define	STRUCT_FROM_LINK(l,t,m) ((t *)((byte *)l - (int)&(((t *)0)->m)))

//============================================================================

#ifndef NULL
#define NULL ((void *)0)
#endif

//============================================================================

extern	qboolean	bigendien;

extern	int		(*BigLong)		(int l);
extern	float	(*BigFloat)		(float l);
extern	short	(*BigShort)		(short l);
extern	int		(*LittleLong)	(int l);
extern	float	(*LittleFloat)	(float l);
extern	short	(*LittleShort)	(short l);

//============================================================================

void	MSG_WriteByte	(sizebuf_t *sb, int c);
void	MSG_WriteShort	(sizebuf_t *sb, int c);
void	MSG_WriteLong	(sizebuf_t *sb, int c);
void	MSG_WriteFloat	(sizebuf_t *sb, float f);
void	MSG_WriteString	(sizebuf_t *sb, char *s);
void	MSG_WriteCoord	(sizebuf_t *sb, float f);
void	MSG_WriteAngle	(sizebuf_t *sb, float f);

extern	int			msg_readcount;
extern	qboolean	msg_badread;		// set if a read goes beyond end of message

float	MSG_ReadFloat		(void);
float	MSG_ReadCoord		(void);
float	MSG_ReadAngle		(void);
int		MSG_ReadByte		(void);
int		MSG_ReadShort		(void);
int		MSG_ReadLong		(void);
char*	MSG_ReadString		(void);
void	MSG_BeginReading	(void);

//============================================================================

void	MR_memset		(void *dest, int fill, int count);
void	MR_memcpy		(void *dest, void *src, int count);
void	MR_strcpy		(char *dest, char *src);
void	MR_strncpy		(char *dest, char *src, int count);
void	MR_strcat		(char *dest, char *src);
int		MR_strlen		(char *str);
int		MR_strcmp		(char *s1, char *s2);
int		MR_strncmp		(char *s1, char *s2, int count);
int		MR_strcasecmp	(char *s1, char *s2);
int		MR_strncasecmp	(char *s1, char *s2, int n);
int		MR_atoi			(char *str);
float	MR_atof			(char *str);

//============================================================================

extern	char		com_token[1024];
extern	qboolean	com_eof;

char *COM_Parse (char *data);


extern	int		com_argc;
extern	char	**com_argv;

int COM_CheckParm (char *parm);
void COM_Init ();
void COM_InitArgv (int argc, char **argv);

void COM_StripExtension (char *in, char *out);
void COM_FileBase (char *in, char *out);

char	*va(char *format, ...);
// does a varargs printf into a temp buffer

//============================================================================

extern int com_filesize;
struct cache_user_s;

extern	char	com_gamedir[MAX_OSPATH];

void COM_WriteFile (char *filename, void *data, int len);
int COM_OpenFile (char *filename, int *hndl);
int COM_FOpenFile (char *filename, FILE **file);
void COM_CloseFile (int h);

byte *COM_LoadStackFile (char *path, void *buffer, int bufsize);
byte *COM_LoadTempFile (char *path);
byte *COM_LoadHunkFile (char *path);
void COM_LoadCacheFile (char *path, struct cache_user_s *cu);

#endif //__COMMON__
