/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include		"../vid.h"
#include		"../quakedef.h"

#define			MAXCMDLINE		256
#define			MAXPRINTMSG		4096
#define			CON_TEXTSIZE	16384
#define			NUM_CON_TIMES	4

extern	int		edit_line;
extern	int		key_linepos;
extern	char	key_lines[32][MAXCMDLINE];		
extern	void	M_f (int state);
extern	void	Con_OpenDebugLog (void);

FILE			*logfile;

float			con_cursorspeed = 4;
float			con_times[NUM_CON_TIMES];

int				con_x;
int				con_current;
int 			con_linewidth;
int				con_totallines;
int				con_backscroll;

char			*con_text=0;

qboolean 		con_forcedup;
qboolean		con_initialized;

/*
===================
Con_ToggleConsole_f
===================
*/
void Con_ToggleConsole_f (void)
{
	if (key_dest == key_console)
	{
		if (cls.state == ca_connected)
		{
			key_dest = key_game;
			key_lines[edit_line][1] = 0;	// clear any typing
			key_linepos = 1;
		}
		else
			M_f (0);
	}
	else
		key_dest = key_console;
	
	SCR_EndLoadingPlaque ();
	MR_memset (con_times, 0, sizeof(con_times));
}

/*
===========
Con_Clear_f
===========
*/
void Con_Clear_f (void)
{
	if (con_text)
		MR_memset (con_text, ' ', CON_TEXTSIZE);
}
						
/*
===============
Con_ClearNotify
===============
*/
void Con_ClearNotify (void)
{
	int		i;
	
	for (i=0; i<NUM_CON_TIMES; i++)
		con_times[i] = 0;
}
						
/*
=================
Con_MessageMode_f
=================
*/
void Con_MessageMode_f (void)
{
	key_dest = key_message;
}
						
/*
===============
Con_CheckResize
===============
*/
void Con_CheckResize (void)
{
	int		i, j, width, oldwidth, oldtotallines, numlines, numchars;
	char	tbuf[CON_TEXTSIZE];

	width = (vid.width / 12);

	if (width == con_linewidth)
		return;

	if (width < 1)			// video hasn't been initialized yet
	{
		width = 38;
		con_linewidth = width;
		con_totallines = CON_TEXTSIZE / con_linewidth;
		MR_memset (con_text, ' ', CON_TEXTSIZE);
	}
	else
	{
		oldwidth = con_linewidth;
		con_linewidth = width;
		oldtotallines = con_totallines;
		con_totallines = CON_TEXTSIZE / con_linewidth;
		numlines = oldtotallines;

		if (con_totallines < numlines)
			numlines = con_totallines;

		numchars = oldwidth;
	
		if (con_linewidth < numchars)
			numchars = con_linewidth;

		MR_memcpy (tbuf, con_text, CON_TEXTSIZE);
		MR_memset (con_text, ' ', CON_TEXTSIZE);

		for (i=0 ; i<numlines ; i++)
			for (j=0 ; j<numchars ; j++)
				con_text[(con_totallines - 1 - i) * con_linewidth + j] = tbuf[((con_current - i + oldtotallines) % oldtotallines) * oldwidth + j];

		Con_ClearNotify ();
	}

	con_backscroll	= 0;
	con_current		= con_totallines - 1;
}

/*
========
Con_Init
========
*/
void Con_Init (void)
{
	con_text = Hunk_AllocName (CON_TEXTSIZE, "context");
	MR_memset (con_text, ' ', CON_TEXTSIZE);
	con_linewidth = -1;
	Con_CheckResize ();
	
	Con_Printf ("Console initialized.\n");

	Cmd_AddCommand ("toggleconsole", Con_ToggleConsole_f);
	Cmd_AddCommand ("messagemode", Con_MessageMode_f);
	Cmd_AddCommand ("clear", Con_Clear_f);
	con_initialized = true;
}

/*
============
Con_Linefeed
============
*/
void Con_Linefeed (void)
{
	con_x = 0;
	con_current++;
	MR_memset (&con_text[(con_current%con_totallines)*con_linewidth], ' ', con_linewidth);
}

/*
=========
Con_Print
=========
*/
void Con_Print (char *txt)
{
	int		y;
	int		c, l;
	static int	cr;
	int		mask;
	
	con_backscroll = 0;

	if (txt[0] == 1)
	{
		mask = 128;
		txt++;
	}
	else if (txt[0] == 2)
	{
		mask = 128;
		txt++;
	}
	else
		mask = 0;

	while ((c = *txt))
	{
		for (l=0 ; l< con_linewidth ; l++)
			if ( txt[l] <= ' ')
				break;

		if (l != con_linewidth && (con_x + l > con_linewidth))
			con_x = 0;

		txt++;

		if (cr)
		{
			con_current--;
			cr = false;
		}

		
		if (!con_x)
		{
			Con_Linefeed ();
		// mark time for transparent overlay
			if (con_current >= 0)
				con_times[con_current % NUM_CON_TIMES] = realtime;
		}

		switch (c)
		{
		case '\n':
			con_x = 0;
			break;

		case '\r':
			con_x = 0;
			cr = 1;
			break;

		default:	// display character and advance
			y = con_current % con_totallines;
			con_text[y*con_linewidth+con_x] = c | mask;
			con_x++;
			if (con_x >= con_linewidth)
				con_x = 0;
			break;
		}
	}
}

/*
================
Con_OpenDebugLog
================
*/
void Con_OpenDebugLog (void)
{
	logfile = fopen("data/MiniRacer.log","w");
}

/*
============
Con_DebugLog
============
*/
void Con_DebugLog (const char *fmt, ...)
{
	va_list		args;

	va_start (args, fmt);

	vfprintf (logfile, fmt, args);

	va_end(args);
}

/*
=================
Con_CloseDebugLog
=================
*/
void Con_CloseDebugLog (void)
{
	fclose (logfile);
}

/*
==========
Con_Printf
==========
*/
void Con_Printf (char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];
	static qboolean	inupdate;
	
	va_start (argptr,fmt);
	_vsnprintf (msg, sizeof(msg), fmt, argptr);
	va_end (argptr);
	
	Sys_Printf ("%s", msg);

	if (!con_initialized)
		return;
		
	if (cls.state == ca_dedicated)
		return;

	Con_Print (msg);
	
	if (cls.signon != SIGNONS && !scr_disabled_for_loading)
	{
		if (!inupdate)
		{
			inupdate = true;
			SCR_UpdateScreen ();
			inupdate = false;
		}
	}
}

/*
===========
Con_DPrintf
===========
*/
void Con_DPrintf (char *fmt, ...)
{
	va_list		argptr;
	char		msg[MAXPRINTMSG];
		
	va_start (argptr,fmt);
	_vsnprintf (msg, sizeof(msg), fmt, argptr);
	va_end (argptr);

	if (developer.ivalue)
		Con_Printf ("%s", msg);
}

/*
==============
Con_SafePrintf
==============
*/
void Con_SafePrintf (char *fmt, ...)
{
	va_list		argptr;
	char		msg[1024];
	int			temp;
		
	va_start (argptr,fmt);
	_vsnprintf (msg, sizeof(msg), fmt, argptr);
	va_end (argptr);

	temp = scr_disabled_for_loading;
	scr_disabled_for_loading = true;
	Con_Printf ("%s", msg);
	scr_disabled_for_loading = temp;
}

/*
=============
Con_DrawInput
=============
*/
void Con_DrawInput (void)
{
	int		y;
	int		i;
	char	*text;

	if (key_dest != key_console && !con_forcedup)
		return;

	text = key_lines[edit_line];
	
	text[key_linepos] = 10+((int)(realtime*con_cursorspeed)&1);
	
	for (i=key_linepos+1 ; i< con_linewidth ; i++)
		text[i] = ' ';
		
	if (key_linepos >= con_linewidth)
		text += 1 + key_linepos - con_linewidth;
		
	y = vid.height-16;

	for (i=0 ; i<con_linewidth ; i++)
		Draw_Character (i * 12, vid.height - 16, text[i], 0, 255, 255, 255, 255);

	key_lines[edit_line][key_linepos] = 0;
}

/*
==============
Con_DrawNotify
==============
*/
void Con_DrawNotify (void)
{
	int		x, v;
	char	*text;
	int		i;
	float	time;
	extern char chat_buffer[];

	v = 0;
	for (i= con_current-NUM_CON_TIMES+1 ; i<=con_current ; i++)
	{
		if (i < 0)
			continue;
		time = con_times[i % NUM_CON_TIMES];
		if (time == 0)
			continue;
		time = realtime - time;
		if (time > 3)
			continue;
		text = con_text + (i % con_totallines)*con_linewidth;
		
		for (x = 0 ; x < con_linewidth ; x++)
			Draw_Character (x * 12, v, text[x], 0, 255, 255, 255, 255);

		v += 16;
	}

	if (key_dest == key_message)
	{
		x = 0;
		
		Draw_String (0, v, "say:", 0, 255, 255, 255, 255);

		while(chat_buffer[x])
		{
			Draw_Character ((x + 4) * 12, v, chat_buffer[x], 0, 255, 255, 255, 255);
			x++;
		}
		Draw_Character ((x + 4) * 12, v, 10+((int)(realtime*con_cursorspeed)&1), 0, 255, 255, 255, 255);
		v += 16;
	}
}

/*
===============
Con_DrawConsole
===============
*/
void Con_DrawConsole (void)
{
	int				i, x, y, j, rows;
	char			*text;
	
	Draw_ConsoleBackground ();

	rows = (vid.height-16)>>4;
	y = vid.height - 16 - (rows<<4);

	for (i= con_current - rows + 1 ; i<=con_current ; i++, y+=16 )
	{
		j = i - con_backscroll;
		if (j<0)
			j = 0;
		text = con_text + (j % con_totallines)*con_linewidth;

		for (x=0 ; x<con_linewidth ; x++)
			Draw_Character (x * 12, y, text[x], 0, 255, 255, 255, 255);
	}

	Con_DrawInput ();
}
