/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include	"../quakedef.h"
#include	"../game.h"

/*
===========
Host_Quit_f
===========
*/

extern void M_Quit_f (void);

void Host_Quit_f (void)
{
	if (key_dest != key_console && cls.state != ca_dedicated)
	{
		M_Quit_f ();
		return;
	}
	CL_Disconnect ();
	Host_ShutdownServer(false);		

	Sys_Quit ();
}

/*
=============
Host_Status_f
=============
*/
void Host_Status_f (void)
{
	client_t	*client;
	int			seconds;
	int			minutes;
	int			hours = 0;
	int			j;
	void		(*print) (char *fmt, ...);
	
	if (cmd_source == src_command)
	{
		if (!sv.active)
		{
			Cmd_ForwardToServer ();
			return;
		}
		print = Con_Printf;
	}
	else
		print = SV_ClientPrintf;

	print ("host:    %s\n", Cvar_VariableString ("hostname"));
	print ("version: %4.2f\n", VERSION);
	if (tcpipAvailable)
		print ("tcp/ip:  %s\n", my_tcpip_address);
	print ("map:     %s\n", sv.name);
	print ("players: %i active (%i max)\n\n", net_activeconnections, svs.maxclients);
	for (j=0, client = svs.clients ; j<svs.maxclients ; j++, client++)
	{
		if (!client->active)
			continue;
		seconds = (int)(net_time - client->netconnection->connecttime);
		minutes = seconds * 0.0166666666;	// Speed... were ( / 60)
		if (minutes)
		{
			seconds -= (minutes * 60);
			hours = minutes * 0.0166666666;	// Speed... were ( / 60)
			if (hours)
				minutes -= (hours * 60);
		}
		else
			hours = 0;
		print ("#%-2u %-16.16s  %2i:%02i:%02i\n", j+1, client->name, hours, minutes, seconds);
		print ("   %s\n", client->netconnection->address);
	}
}

/*
===========
Host_Ping_f
===========
*/
void Host_Ping_f (void)
{
	int		i, j;
	float	total;
	client_t	*client;
	
	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	SV_ClientPrintf ("Client ping time:\n");
	for (i=0, client = svs.clients ; i<svs.maxclients ; i++, client++)
	{
		if (!client->active)
			continue;
		total = 0;
		for (j=0 ; j<NUM_PING_TIMES ; j++)
			total+=client->ping_times[j];
		total /= NUM_PING_TIMES;
		SV_ClientPrintf ("%4i %s\n", (int)(total*1000), client->name);
	}
}

/*
===============================================================================

SERVER TRANSITIONS

===============================================================================
*/


/*
==========
Host_Map_f
==========
*/
extern int	startmap;
void R_LoadSkyBox (void);

void Host_Map_f (void)
{
	int		i;
	char	name[MAX_QPATH];

	if (cmd_source != src_command)
		return;

	Draw_ConsoleBackground ();
	GL_EndRendering ();

	if (key_dest != key_menu)
		startmap = 0;

	cls.demonum = -1;

	CL_Disconnect ();
	Host_ShutdownServer(false);		

	MR_memset (&miniracer, 0, sizeof(miniracer));

	key_dest = key_game;
	SCR_BeginLoadingPlaque ();

	cls.mapstring[0] = 0;
	for (i=0 ; i<Cmd_Argc() ; i++)
	{
		MR_strcat (cls.mapstring, Cmd_Argv(i));
		MR_strcat (cls.mapstring, " ");
	}
	MR_strcat (cls.mapstring, "\n");

	MR_strcpy (name, Cmd_Argv(1));

	SV_SpawnServer (name);

	if (!sv.active)
		return;
	
	if (cls.state != ca_dedicated)
		Cmd_ExecuteString ("connect local", src_command);

	R_LoadSkyBox ();
}

/*
==================
Host_Changelevel_f
==================
*/
void Host_Changelevel_f (void)
{
	char	mapname[MAX_QPATH];

	if (Cmd_Argc() != 2)
	{
		Con_Printf ("changelevel <levelname> : continue game on a new level\n");
		return;
	}
	if (!sv.active || cls.demoplayback)
	{
		Con_Printf ("Only the server may changelevel\n");
		return;
	}

	miniracer.started = false;

	MR_strcpy (mapname, Cmd_Argv(1));
	SV_SpawnServer (mapname);
}

/*
==============
Host_Restart_f
==============
*/
void Host_Restart_f (void)
{
	char	mapname[MAX_QPATH];

	if (cls.demoplayback || !sv.active)
		return;

	if (cmd_source != src_command)
		return;

	miniracer.started = false;

	MR_strcpy (mapname, sv.name);
	SV_SpawnServer (mapname);
}

/*
================
Host_Reconnect_f
================
*/
void Host_Reconnect_f (void)
{
	SCR_BeginLoadingPlaque ();
	cls.signon = 0;
}

/*
==============
Host_Connect_f
==============
*/
void Host_Connect_f (void)
{
	char	name[MAX_QPATH];
	
	cls.demonum = -1;
	if (cls.demoplayback)
	{
		CL_StopPlayback ();
		CL_Disconnect ();
	}
	MR_strcpy (name, Cmd_Argv(1));
	CL_EstablishConnection (name);
	Host_Reconnect_f ();
}

//============================================================================

/*
===========
Host_Name_f
===========
*/
void Host_Name_f (void)
{
	char	*newName;

	if (Cmd_Argc () == 1)
	{
		Con_Printf ("\"name\" is \"%s\"\n", _cl_name.string);
		return;
	}
	if (Cmd_Argc () == 2)
		newName = Cmd_Argv(1);	
	else
		newName = Cmd_Args();
	newName[15] = 0;

	if (cmd_source == src_command)
	{
		if (MR_strcmp (_cl_name.string, newName) == 0)
			return;
		Cvar_Set ("_cl_name", newName);
		if (cls.state == ca_connected)
			Cmd_ForwardToServer ();
		return;
	}

	if (host_client->name[0] && MR_strcmp (host_client->name, "unconnected") )
		if (MR_strcmp (host_client->name, newName) != 0)
			Con_Printf ("%s renamed to %s\n", host_client->name, newName);
	MR_strcpy (host_client->name, newName);

	_snprintf (host_client->edict->game.netname, sizeof(host_client->edict->game.netname), host_client->name);
	
// send notification to all clients
	
	MSG_WriteByte (&sv.reliable_datagram, svc_updatename);
	MSG_WriteByte (&sv.reliable_datagram, host_client - svs.clients);
	MSG_WriteString (&sv.reliable_datagram, host_client->name);
}

/*
==========
Host_Say_f
==========
*/
void Host_Say_f(void)
{
	client_t *client;
	client_t *save;
	int		j;
	char	*p;
	unsigned char	text[64];
	qboolean	fromServer = false;

	if (cmd_source == src_command)
	{
		if (cls.state == ca_dedicated)
			fromServer = true;
		else
		{
			Cmd_ForwardToServer ();
			return;
		}
	}

	if (Cmd_Argc () < 2)
		return;

	save = host_client;

	p = Cmd_Args();
// remove quotes if present
	if (*p == '"')
	{
		p++;
		p[MR_strlen (p)-1] = 0;
	}

// turn on color set 1
	if (!fromServer)
		_snprintf (text, sizeof(text), "%c%s: ", 1, save->name);
	else
		_snprintf (text, sizeof(text), "%c<%s> ", 1, hostname.string);

	j = sizeof(text) - 2 - MR_strlen (text);  // -2 for /n and null terminator
	if (MR_strlen (p) > j)
		p[j] = 0;

	MR_strcat (text, p);
	MR_strcat (text, "\n");

	for (j = 0, client = svs.clients; j < svs.maxclients; j++, client++)
	{
		if (!client || !client->active || !client->spawned)
			continue;
		host_client = client;
		SV_ClientPrintf("%s", text);
	}
	host_client = save;

	Sys_Printf("%s", &text[1]);
}

/*
===========
Host_Tell_f
===========
*/
void Host_Tell_f(void)
{
	client_t *client;
	client_t *save;
	int		j;
	char	*p;
	char	text[64];

	if (cmd_source == src_command)
	{
		Cmd_ForwardToServer ();
		return;
	}

	if (Cmd_Argc () < 3)
		return;

	MR_strcpy (text, host_client->name);
	MR_strcat (text, ": ");

	p = Cmd_Args();

// remove quotes if present
	if (*p == '"')
	{
		p++;
		p[MR_strlen (p)-1] = 0;
	}

// check length & truncate if necessary
	j = sizeof(text) - 2 - MR_strlen (text);  // -2 for /n and null terminator
	if (MR_strlen (p) > j)
		p[j] = 0;

	MR_strcat (text, p);
	MR_strcat (text, "\n");

	save = host_client;
	for (j = 0, client = svs.clients; j < svs.maxclients; j++, client++)
	{
		if (!client->active || !client->spawned)
			continue;
		if (MR_strcasecmp(client->name, Cmd_Argv(1)))
			continue;
		host_client = client;
		SV_ClientPrintf("%s", text);
		break;
	}
	host_client = save;
}

/*
===============
Host_Carcolor_f
===============
*/
void Host_Carcolor_f (void)
{
	byte	hostcolor[3];
	
	if (Cmd_Argc() == 2)
		hostcolor[0] = hostcolor[1] = hostcolor[2] = MR_atoi (Cmd_Argv(1));	
	else if (Cmd_Argc() == 4)
	{
		hostcolor[0]	= MR_atoi (Cmd_Argv(1));
		hostcolor[1]	= MR_atoi (Cmd_Argv(2));
		hostcolor[2]	= MR_atoi (Cmd_Argv(3));
	}
	else
		return;

	if (cmd_source == src_command)
	{
		Cvar_SetValue ("_cl_color_red",	hostcolor[0]);
		Cvar_SetValue ("_cl_color_green",hostcolor[1]);
		Cvar_SetValue ("_cl_color_blue",	hostcolor[2]);

		if (cls.state == ca_connected)
			Cmd_ForwardToServer ();
		return;
	}

	host_client->carcolors[0] = hostcolor[0];
	host_client->carcolors[1] = hostcolor[1];
	host_client->carcolors[2] = hostcolor[2];

	MSG_WriteByte (&sv.reliable_datagram, svc_updatecolors);
	MSG_WriteByte (&sv.reliable_datagram, host_client - svs.clients);
	MSG_WriteByte (&sv.reliable_datagram, hostcolor[0]);
	MSG_WriteByte (&sv.reliable_datagram, hostcolor[1]);
	MSG_WriteByte (&sv.reliable_datagram, hostcolor[2]);
}

/*
===============
Host_Carmodel_f
===============
*/
void Host_Carmodel_f (void)
{
	char	hostmodel[32];
	
	if (Cmd_Argc() != 2)
		return;

	MR_strcpy (hostmodel, Cmd_Argv(1));

	if (cmd_source == src_command)
	{
		if (cls.state == ca_connected)
			Cmd_ForwardToServer ();
		return;
	}
	_snprintf (host_client->carmodel, sizeof(host_client->carmodel), "%s.car", hostmodel);
}

/*
===============
Host_PreSpawn_f
===============
*/
void Host_PreSpawn_f (void)
{
	if (cmd_source == src_command)
	{
		Con_Printf ("prespawn is not valid from the console\n");
		return;
	}

	if (host_client->spawned)
	{
		Con_Printf ("prespawn not valid -- already spawned\n");
		return;
	}
	
	SZ_Write (&host_client->message, sv.signon.data, sv.signon.cursize);
	MSG_WriteByte (&host_client->message, svc_signonnum);
	MSG_WriteByte (&host_client->message, 2);
	host_client->sendsignon = true;
}

/*
============
Host_Spawn_f
============
*/
void Host_Spawn_f (void)
{
	client_t	*client;
	edict_t		*ent;
	int			i;

	if (cmd_source == src_command)
	{
		Con_Printf ("spawn is not valid from the console\n");
		return;
	}

	if (host_client->spawned)
	{
		Con_Printf ("Spawn not valid -- already spawned\n");
		return;
	}

	// set up the edict
	ent = host_client->edict;

	if (miniracer.started)
		BotDisconnect (ent);

	MR_memset (&ent->v,		0, sizeof(ent->v));
	MR_memset (&ent->game,	0, sizeof(ent->game));

	_snprintf (ent->game.netname, sizeof(ent->game.netname), host_client->name);

	VectorCopy (host_client->carcolors, ent->game.color);
	ent->v.model = host_client->carmodel;

	miniracer.self = sv_player;

	ClientConnect (sv_player);

	if (miniracer.started)
		PutObserverInServer (sv_player);
	else
		PutClientInServer (sv_player);

	SZ_Clear (&host_client->message);

	MSG_WriteByte (&host_client->message, svc_time);
	MSG_WriteFloat (&host_client->message, sv.time);

	for (i=0, client = svs.clients ; i<svs.maxclients ; i++, client++)
	{
		MSG_WriteByte	(&host_client->message, svc_updatename);
		MSG_WriteByte	(&host_client->message, i);
		MSG_WriteString	(&host_client->message, client->name);
		MSG_WriteByte	(&host_client->message, svc_updatecolors);
		MSG_WriteByte	(&host_client->message, i);
		MSG_WriteByte	(&host_client->message, client->carcolors[0]);
		MSG_WriteByte	(&host_client->message, client->carcolors[1]);
		MSG_WriteByte	(&host_client->message, client->carcolors[2]);
		MSG_WriteByte	(&host_client->message, svc_updatetime);
		MSG_WriteByte	(&host_client->message, i);
		MSG_WriteFloat	(&host_client->message, 3599.999f);
		MSG_WriteFloat	(&host_client->message, 3599.999f);
		MSG_WriteByte	(&host_client->message, svc_updatelaps);
		MSG_WriteByte	(&host_client->message, i);
		MSG_WriteByte	(&host_client->message, 0);
	}

	MSG_WriteByte (&host_client->message, svc_signonnum);
	MSG_WriteByte (&host_client->message, 3);
	host_client->sendsignon = true;
}

/*
============
Host_Begin_f
============
*/
void Host_Begin_f (void)
{
	if (cmd_source == src_command)
	{
		Con_Printf ("begin is not valid from the console\n");
		return;
	}

	host_client->spawned = true;
}

/*
===============================================================================

DEMO LOOP CONTROL

===============================================================================
*/

/*
=================
Host_Startdemos_f
=================
*/
void Host_Startdemos_f (void)
{
	int		i, c;

	if (cls.state == ca_dedicated)
	{
		if (!sv.active)
			Cbuf_AddText ("map start\n");
		return;
	}

	if (sv.active)
		return;

	c = Cmd_Argc() - 1;
	if (c > MAX_DEMOS)
	{
		Con_Printf ("Max %i demos in demoloop\n", MAX_DEMOS);
		c = MAX_DEMOS;
	}
	Con_Printf ("%i demo(s) in loop\n", c);

	for (i=1 ; i<c+1 ; i++)
		MR_strncpy (cls.demos[i-1], Cmd_Argv(i), sizeof(cls.demos[0])-1);

	if (cls.demonum != -1 && !cls.demoplayback)
	{
		cls.demonum = 0;
		CL_NextDemo ();
	}
	else
		cls.demonum = -1;
}

/*
============
Host_Demos_f
============
*/
void Host_Demos_f (void)
{
	if (cls.state == ca_dedicated)
		return;
	if (cls.demonum == -1)
		cls.demonum = 1;
	CL_Disconnect_f ();
	CL_NextDemo ();
}

/*
===============
Host_Stopdemo_f
===============
*/
void Host_Stopdemo_f (void)
{
	if (cls.state == ca_dedicated)
		return;
	if (!cls.demoplayback)
		return;
	CL_StopPlayback ();
	CL_Disconnect ();
}

//=============================================================================

/*
=================
Host_InitCommands
=================
*/
void Host_InitCommands (void)
{
	Cmd_AddCommand ("map",			Host_Map_f);
	Cmd_AddCommand ("say",			Host_Say_f);
	Cmd_AddCommand ("name",			Host_Name_f);
	Cmd_AddCommand ("ping",			Host_Ping_f);
	Cmd_AddCommand ("quit",			Host_Quit_f);
	Cmd_AddCommand ("tell",			Host_Tell_f);
	Cmd_AddCommand ("begin",		Host_Begin_f);
	Cmd_AddCommand ("demos",		Host_Demos_f);
	Cmd_AddCommand ("spawn",		Host_Spawn_f);
	Cmd_AddCommand ("status",		Host_Status_f);
	Cmd_AddCommand ("connect",		Host_Connect_f);
	Cmd_AddCommand ("restart",		Host_Restart_f);
	Cmd_AddCommand ("prespawn",		Host_PreSpawn_f);
	Cmd_AddCommand ("stopdemo",		Host_Stopdemo_f);
	Cmd_AddCommand ("_carcolor",	Host_Carcolor_f);
	Cmd_AddCommand ("_carmodel",	Host_Carmodel_f);
	Cmd_AddCommand ("reconnect",	Host_Reconnect_f);
	Cmd_AddCommand ("startdemos",	Host_Startdemos_f);
	Cmd_AddCommand ("changelevel",	Host_Changelevel_f);
}
