/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __MATHLIB__
#define __MATHLIB__

typedef float vec3_t[3];

struct	mplane_s;

extern	vec3_t vec3_origin;

double	MR_sin (double x);
double	MR_cos (double x);
double	MR_pow (double x, double y);
double	MR_asin (double x);
double	MR_ceil (double x);
double	MR_fabs (double x);
double	MR_floor (double x);
double	MR_atan2 (double x, double y);

float	DotProduct (float v1[3], float v2[3]); 
float	MR_sqrt (float n);

int		MR_rand ();

void	VectorNormalize (vec3_t v);
void	AngleVectors (vec3_t angles, vec3_t forward, vec3_t right, vec3_t up);
void	BoxOnPlaneSideClassify(struct mplane_s *p);

#define bound(min,num,max) (num >= min ? (num < max ? num : max) : min)

#define DEG2RAD(a) ((a) * (3.14159265358979323846f / 180.0f))
#define RAD2DEG(a) ((a) * (180.0f / 3.14159265358979323846f))

#define ANGLEMOD(a) (((int) ((a) * (65536.0f / 360.0f)) & 65535) * (360.0f / 65536.0f))
#define PlaneDiff(point,plane) (((plane)->type < 3 ? (point)[(plane)->type] : DotProduct((point), (plane)->normal)) - (plane)->dist)
#define Length(v) (MR_sqrt(DotProduct(v,v)))
#define VectorCompare(a,b) ((a[0] == b[0]) & (a[1] == b[1]) & (a[2] == b[2]))

#define VectorSubtract(a,b,c)					\
(												\
	c[0]	= a[0] - b[0],						\
	c[1]	= a[1] - b[1],						\
	c[2]	= a[2] - b[2]						\
)

#define VectorAdd(a,b,c)						\
(												\
	c[0]	= a[0] + b[0],						\
	c[1]	= a[1] + b[1],						\
	c[2]	= a[2] + b[2]						\
)

#define VectorCopy(a,b)							\
(												\
	b[0]	= a[0],								\
	b[1]	= a[1],								\
	b[2]	= a[2]								\
)

#define VectorMA(a,s,b,c)						\
(												\
	c[0]	= a[0] + (s * b[0]),				\
	c[1]	= a[1] + (s * b[1]),				\
	c[2]	= a[2] + (s * b[2])					\
)

#define CrossProduct(v1,v2,cross)				\
(												\
	cross[0] = v1[1] * v2[2] - v1[2] * v2[1],	\
	cross[1] = v1[2] * v2[0] - v1[0] * v2[2],	\
	cross[2] = v1[0] * v2[1] - v1[1] * v2[0]	\
)

#define VectorInverse(v)						\
(												\
	v[0]	= -v[0],							\
	v[1]	= -v[1],							\
	v[2]	= -v[2]								\
)

#define VectorScale(a,b,c)						\
(												\
	c[0]	= a[0] * b,							\
	c[1]	= a[1] * b,							\
	c[2]	= a[2] * b							\
)

#define VectorClear(a)							\
(												\
	a[0]	= 0,								\
	a[1]	= 0,								\
	a[2]	= 0									\
)

#define VectorNegate(a,b)						\
(												\
	b[0]	= -a[0],							\
	b[1]	= -a[1],							\
	b[2]	= -a[2]								\
)

#define BOX_ON_PLANE_SIDE(emins, emaxs, p)		\
	(((p)->type < 3)?							\
	(											\
		((p)->dist <= (emins)[(p)->type])?		\
			1									\
		:										\
		(										\
			((p)->dist >= (emaxs)[(p)->type])?	\
				2								\
			:									\
				3								\
		)										\
	)											\
	:											\
		(p)->BoxOnPlaneSideFunc( (emins), (emaxs), (p)))

#endif //__MATHLIB__
