/*
Copyright (C) 2001-2002 Tomas Jakobsson.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"

#define		NUM_CAMS	8

extern	camera_t	camera;

typedef struct
{
	void	(*func)(void);
	char	*name;
	byte	interp;
} cameramode_t;

void	CameraMode0	(void);
void	CameraMode1	(void);
void	CameraMode2	(void);
void	CameraMode3	(void);
void	CameraMode4	(void);
void	CameraMode5	(void);
void	CameraMode6	(void);
void	CameraMode7	(void);

cameramode_t cameramode[NUM_CAMS] =
{
	{CameraMode0, "Sky Cam", 25},
	{CameraMode1, "Chase Cam", 0},
	{CameraMode2, "TV Cam", 25},
	{CameraMode3, "TV Cam with Zoom", 25},
	{CameraMode4, "User Sky Cam", 25},
	{CameraMode5, "User Chase Cam", 0},
	{CameraMode6, "User Chase Cam with auto angling", 0},
	{CameraMode7, "User TV Cam", 25}
};

cvar_t	cam_forward	= {"cam_forward",	"0",	true};
cvar_t	cam_right	= {"cam_right",		"0",	true};
cvar_t	cam_up		= {"cam_up",		"0",	true};
cvar_t	cam_pitch	= {"cam_pitch",		"0",	true};
cvar_t	cam_roll	= {"cam_roll",		"0",	true};
cvar_t	cam_yaw		= {"cam_yaw",		"0",	true};
cvar_t	cam_mode	= {"cam_mode",		"0",	true};
cvar_t	cam_interp	= {"cam_interp",	"0",	true};

/*
========
CameraUp
========
*/
void CameraUp (void)
{
	cam_mode.ivalue++;

	if (cam_mode.ivalue == NUM_CAMS)
		cam_mode.ivalue = 0;

	Cvar_SetValue ("cam_mode", cam_mode.ivalue);

	Con_Printf ("%s\n", cameramode[cam_mode.ivalue].name);
}

/*
==========
CameraDown
==========
*/
void CameraDown (void)
{
	cam_mode.ivalue--;

	if (cam_mode.ivalue == -1)
		cam_mode.ivalue = NUM_CAMS - 1;

	Cvar_SetValue ("cam_mode", cam_mode.ivalue);

	Con_Printf ("%s\n", cameramode[cam_mode.ivalue].name);
}

/*
========
ClientUp
========
*/
void ClientUp (void)
{
	do
	{
		cl.viewentity++;

		if (cl.viewentity > 8)
			cl.viewentity = 1;

	} while (!cl.times[cl.viewentity - 1].name[0]);

	Con_Printf ("Following %s\n", 	cl.times[cl.viewentity - 1].name);
}

/*
==========
ClientDown
==========
*/
void ClientDown (void)
{
	do
	{
		cl.viewentity--;
	
		if (cl.viewentity < 1)
			cl.viewentity = 8;

	} while (!cl.times[cl.viewentity - 1].name[0]);

	Con_Printf ("Following %s\n", 	cl.times[cl.viewentity - 1].name);
}

/*
===========
CameraMode0
===========
*/
void CameraMode0 (void)
{
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	camera.dest_origin[0] = ent->origin[0] - 80;
	camera.dest_origin[1] = ent->origin[1] - 80;
	camera.dest_origin[2] = ent->origin[2] + 200;

	camera.dest_angles[0] = 60;
	camera.dest_angles[1] = 45;
	camera.dest_angles[2] = 0;
}

/*
===========
CameraMode1
===========
*/
void CameraMode1 (void)
{
	vec3_t		forward, right, up;
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	AngleVectors (ent->angles, forward, right, up);

	VectorScale (forward, -25, forward);
	VectorScale (up, 10, up);

	camera.dest_origin[0] = ent->origin[0] + forward[0] + up[0];
	camera.dest_origin[1] = ent->origin[1] + forward[1] + up[1];
	camera.dest_origin[2] = ent->origin[2] + forward[2] + up[2];

	camera.dest_angles[0] = -ent->angles[0];
	camera.dest_angles[1] =  ent->angles[1];
	camera.dest_angles[2] =  ent->angles[2];
}

/*
===========
CameraMode2
===========
*/
void CameraMode2 (void)
{
	vec3_t		dist;
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	camera.dest_origin[0] = 0;
	camera.dest_origin[1] = 0;
	camera.dest_origin[2] = 150;

	VectorSubtract (ent->origin, camera.dest_origin, dist);
	VectorNormalize (dist);

	camera.dest_angles[0] = -(RAD2DEG(MR_asin (dist[2])));
	camera.dest_angles[1] = RAD2DEG(MR_atan2 (dist[1], dist[0]));
	camera.dest_angles[2] = 0;
}

/*
===========
CameraMode3
===========
*/
void CameraMode3 (void)
{
	vec3_t		dist;
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	camera.dest_origin[0] = 0;
	camera.dest_origin[1] = 0;
	camera.dest_origin[2] = 150;

	VectorSubtract (ent->origin, camera.dest_origin, dist);
	VectorNormalize (dist);

	camera.dest_angles[0] = -(RAD2DEG(MR_asin (dist[2])));
	camera.dest_angles[1] = RAD2DEG(MR_atan2 (dist[1], dist[0]));
	camera.dest_angles[2] = 0;

	VectorSubtract (camera.dest_origin, ent->origin, dist);
	VectorNormalize (dist);

	camera.dest_origin[0] = ent->origin[0] + (dist[0] * 150);
	camera.dest_origin[1] = ent->origin[1] + (dist[1] * 150);
	camera.dest_origin[2] = ent->origin[2] + (dist[2] * 150);
}

/*
===========
CameraMode4
===========
*/
void CameraMode4 (void)
{
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	camera.dest_origin[0] = ent->origin[0] + cam_right.ivalue;
	camera.dest_origin[1] = ent->origin[1] + cam_forward.ivalue;
	camera.dest_origin[2] = ent->origin[2] + cam_up.ivalue;

	camera.dest_angles[0] = cam_pitch.ivalue;
	camera.dest_angles[1] = cam_yaw.ivalue;
	camera.dest_angles[2] = cam_roll.ivalue;
}

/*
===========
CameraMode5
===========
*/
void CameraMode5 (void)
{
	vec3_t		forward, right, up;
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	AngleVectors (ent->angles, forward, right, up);

	camera.dest_origin[0] = ent->origin[0] + forward[0] * cam_forward.ivalue + right[0] * cam_right.ivalue + up[0] * cam_up.ivalue;
	camera.dest_origin[1] = ent->origin[1] + forward[1] * cam_forward.ivalue + right[1] * cam_right.ivalue + up[1] * cam_up.ivalue;
	camera.dest_origin[2] = ent->origin[2] + forward[2] * cam_forward.ivalue + right[2] * cam_right.ivalue + up[2] * cam_up.ivalue;

	camera.dest_angles[0] = -ent->angles[0] + cam_pitch.ivalue;
	camera.dest_angles[1] =  ent->angles[1] + cam_yaw.ivalue;
	camera.dest_angles[2] =  ent->angles[2] + cam_roll.ivalue;
}

/*
===========
CameraMode6
===========
*/
void CameraMode6 (void)
{
	vec3_t		dist;
	vec3_t		forward, right, up;
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	AngleVectors (ent->angles, forward, right, up);

	camera.dest_origin[0] = ent->origin[0] + forward[0] * cam_forward.ivalue + right[0] * cam_right.ivalue + up[0] * cam_up.ivalue;
	camera.dest_origin[1] = ent->origin[1] + forward[1] * cam_forward.ivalue + right[1] * cam_right.ivalue + up[1] * cam_up.ivalue;
	camera.dest_origin[2] = ent->origin[2] + forward[2] * cam_forward.ivalue + right[2] * cam_right.ivalue + up[2] * cam_up.ivalue;

	VectorSubtract (ent->origin, camera.dest_origin, dist);
	VectorNormalize (dist);

	camera.dest_angles[0] = -(RAD2DEG(MR_asin (dist[2])));
	camera.dest_angles[1] = RAD2DEG(MR_atan2 (dist[1], dist[0]));
	camera.dest_angles[2] = 0;
}

/*
===========
CameraMode7
===========
*/
void CameraMode7 (void)
{
	vec3_t		dist;
	entity_t	*ent;

	ent	= &cl_entities[cl.viewentity];

	camera.dest_origin[0] = cam_right.ivalue;
	camera.dest_origin[1] = cam_forward.ivalue;
	camera.dest_origin[2] = cam_up.ivalue;

	VectorSubtract (ent->origin, camera.dest_origin, dist);
	VectorNormalize (dist);

	camera.dest_angles[0] = -(RAD2DEG(MR_asin (dist[2])));
	camera.dest_angles[1] = RAD2DEG(MR_atan2 (dist[1], dist[0]));
	camera.dest_angles[2] = 0;
}

/*
===========
SetupCamera
===========
*/
void SetupCamera (void)
{
	float		cameratime;
	float		temp[3];

	cameramode[cam_mode.ivalue].func ();

	if (!cam_interp.ivalue && !cameramode[cam_mode.ivalue].interp)
	{
		camera.origin[0]	= camera.dest_origin[0];
		camera.origin[1]	= camera.dest_origin[1];
		camera.origin[2]	= camera.dest_origin[2];

		camera.angles[0]	= camera.dest_angles[0];
		camera.angles[1]	= camera.dest_angles[1];
		camera.angles[2]	= camera.dest_angles[2];

		return;
	}

	if (!cam_interp.ivalue)
		cameratime = ((50-cameramode[cam_mode.ivalue].interp)*0.2) * host_frametime;
	else
		cameratime = ((50-cam_interp.ivalue)*0.2) * host_frametime;

	camera.old_origin[0] += (camera.dest_origin[0] - camera.old_origin[0]) * cameratime;
	camera.old_origin[1] += (camera.dest_origin[1] - camera.old_origin[1]) * cameratime;
	camera.old_origin[2] += (camera.dest_origin[2] - camera.old_origin[2]) * cameratime;

	if (camera.old_origin[2] > 0)
		camera.old_origin[2] = 0;

	while (camera.dest_angles[0] < 0)   camera.dest_angles[0] += 360;
	while (camera.dest_angles[0] > 360) camera.dest_angles[0] -= 360;
	while (camera.dest_angles[1] < 0)   camera.dest_angles[1] += 360;
	while (camera.dest_angles[1] > 360) camera.dest_angles[1] -= 360;
	while (camera.dest_angles[2] < 0)   camera.dest_angles[2] += 360;
	while (camera.dest_angles[2] > 360) camera.dest_angles[2] -= 360;

	temp[0] = camera.dest_angles[0] - camera.old_angles[0];
	temp[1] = camera.dest_angles[1] - camera.old_angles[1];
	temp[2] = camera.dest_angles[2] - camera.old_angles[2];

	while (temp[0] < -180) temp[0] += 360;
	while (temp[0] >  180) temp[0] -= 360;
	while (temp[1] < -180) temp[1] += 360;
	while (temp[1] >  180) temp[1] -= 360;
	while (temp[2] < -180) temp[2] += 360;
	while (temp[2] >  180) temp[2] -= 360;

	camera.old_angles[0] += temp[0] * cameratime;
	camera.old_angles[1] += temp[1] * cameratime;
	camera.old_angles[2] += temp[2] * cameratime;

	while (camera.old_angles[0] < 0)   camera.old_angles[0] += 360;
	while (camera.old_angles[0] > 360) camera.old_angles[0] -= 360;
	while (camera.old_angles[1] < 0)   camera.old_angles[1] += 360;
	while (camera.old_angles[1] > 360) camera.old_angles[1] -= 360;
	while (camera.old_angles[2] < 0)   camera.old_angles[2] += 360;
	while (camera.old_angles[2] > 360) camera.old_angles[2] -= 360;

	camera.origin[0]	= camera.old_origin[0];
	camera.origin[1]	= camera.old_origin[1];
	camera.origin[2]	= camera.old_origin[2];

	camera.angles[0]	= camera.old_angles[0];
	camera.angles[1]	= camera.old_angles[1];
	camera.angles[2]	= camera.old_angles[2];
}

/*
===============
View_RenderView
===============
*/

void View_RenderView (void)
{
	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	SetupCamera ();

	R_RenderView ();		
}

/*
===========
Camera_Init
===========
*/
void Camera_Init (void)
{
	Cvar_RegisterVariable (&cam_forward);
	Cvar_RegisterVariable (&cam_right);
	Cvar_RegisterVariable (&cam_up);
	Cvar_RegisterVariable (&cam_pitch);
	Cvar_RegisterVariable (&cam_yaw);
	Cvar_RegisterVariable (&cam_roll);
	Cvar_RegisterVariable (&cam_mode);
	Cvar_RegisterVariable (&cam_interp);

	Cmd_AddCommand ("cameraup",	  CameraUp);
	Cmd_AddCommand ("cameradown", CameraDown);

	Cmd_AddCommand ("clientup",	  ClientUp);
	Cmd_AddCommand ("clientdown", ClientDown);
}
