/*
Copyright (C) 2001-2002 Tomas Jakobsson.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"
#include	"../gl_car.h"
#include	"../menu.h"

extern	int		carskins[8];
extern	char	loadname[32];
extern	void	ColorSkin(int carnum);
extern	byte	lightcolor[3];

#define VectorLittleFloat(a,b)					\
(												\
	b[0]	= LittleFloat(a[0]),				\
	b[1]	= LittleFloat(a[1]),				\
	b[2]	= LittleFloat(a[2])					\
)

#define VectorLittleLong(a,b)					\
(												\
	b[0]	= LittleLong(a[0]),					\
	b[1]	= LittleLong(a[1]),					\
	b[2]	= LittleLong(a[2])					\
)

/*
========
LoadBody
========
*/
void LoadBody (CAR_Part_Body_t *Part, CAR_Part_Body_t *temp)
{
	int	i;

	VectorLittleFloat (temp->Origin, Part->Origin);

	Part->NumVertices	= LittleLong (temp->NumVertices);
	Part->NumTriangles	= LittleLong (temp->NumTriangles);

	for (i=0; i<Part->NumVertices; i++)
	{
		VectorLittleFloat (temp->Verts[i], Part->Verts[i]);
	}

	for (i=0; i<Part->NumTriangles; i++)
	{
		VectorLittleLong  (temp->Tris[i].I, Part->Tris[i].I);
		VectorLittleFloat (temp->Tris[i].U, Part->Tris[i].U);
		VectorLittleFloat (temp->Tris[i].V, Part->Tris[i].V);
	}
}

/*
=========
LoadWheel
=========
*/
void LoadWheel (CAR_Part_Wheel_t *Part, CAR_Part_Wheel_t *temp)
{
	int	i;

	VectorLittleFloat (temp->Origin, Part->Origin);

	Part->NumVertices	= LittleLong (temp->NumVertices);
	Part->NumTriangles	= LittleLong (temp->NumTriangles);

	for (i=0; i<Part->NumVertices; i++)
	{
		VectorLittleFloat (temp->Verts[i], Part->Verts[i]);
		VectorLittleFloat (temp->Norms[i], Part->Norms[i]);
	}

	for (i=0; i<Part->NumTriangles; i++)
	{
		VectorLittleLong  (temp->Tris[i].I, Part->Tris[i].I);
		VectorLittleFloat (temp->Tris[i].U, Part->Tris[i].U);
		VectorLittleFloat (temp->Tris[i].V, Part->Tris[i].V);
	}
}

/*
================
Mod_LoadCARModel
================
*/
void Mod_LoadCARModel (model_t *mod, void *buffer)
{
	int				i, c;
	CAR_Header_t	*header, *temp;
//	CAR_Header2_t	*header2;

	mod->type = mod_car;

	for (i=0 ; i<32 ; i++ )
	{
		c = loadname[i];
		if (!c)
			break;
			
		if (c >= 'a' && c <= 'z')
			c -= ('a' - 'A');
		mod->bigname[i] = c;
	}

	temp = (CAR_Header_t *)buffer;

	header = malloc(sizeof(CAR_Header_t));

	MR_memset (header,	0, sizeof(CAR_Header_t));

	header->Version = LittleLong (temp->Version);

	LoadBody  (&header->Body,     &temp->Body);
	LoadWheel (&header->Wheel[0], &temp->Wheel[0]);
	LoadWheel (&header->Wheel[1], &temp->Wheel[1]);
	LoadWheel (&header->Wheel[2], &temp->Wheel[2]);
	LoadWheel (&header->Wheel[3], &temp->Wheel[3]);

	mod->mins[0]	= -12;
	mod->mins[1]	= -12;
	mod->mins[2]	= -2;
	mod->maxs[0]	= 12;
	mod->maxs[1]	= 12;
	mod->maxs[2]	= 4;

	mod->carheader	= header;
}

/*
===========
GL_DrawBody
===========
*/
void GL_DrawBody (CAR_Part_Body_t *Part)
{
	int	i;

	glPushMatrix();
	glTranslatef (Part->Origin[0], Part->Origin[1], Part->Origin[2]);

	glVertexPointer (3, GL_FLOAT, sizeof(float[3]), Part->Verts);

	glBegin(GL_TRIANGLES);
	for (i=0; i<Part->NumTriangles; i++)
	{
		glTexCoord2f  (Part->Tris[i].U[0], Part->Tris[i].V[0]);
		glArrayElement(Part->Tris[i].I[0]);

		glTexCoord2f  (Part->Tris[i].U[1], Part->Tris[i].V[1]);
		glArrayElement(Part->Tris[i].I[1]);

		glTexCoord2f  (Part->Tris[i].U[2], Part->Tris[i].V[2]);
		glArrayElement(Part->Tris[i].I[2]);
	}
	glEnd();

	glPopMatrix();
}

/*
============
GL_DrawWheel
============
*/
void GL_DrawWheel (CAR_Part_Wheel_t *Part)
{
	int	i;

	glPushMatrix();
	glTranslatef (Part->Origin[0], Part->Origin[1], Part->Origin[2]);

	glVertexPointer (3, GL_FLOAT, sizeof(float[3]), Part->Verts);

	glBegin(GL_TRIANGLES);
	for (i=0; i<Part->NumTriangles; i++)
	{
		glTexCoord2f  (Part->Tris[i].U[0], Part->Tris[i].V[0]);
		glArrayElement(Part->Tris[i].I[0]);

		glTexCoord2f  (Part->Tris[i].U[1], Part->Tris[i].V[1]);
		glArrayElement(Part->Tris[i].I[1]);

		glTexCoord2f  (Part->Tris[i].U[2], Part->Tris[i].V[2]);
		glArrayElement(Part->Tris[i].I[2]);
	}
	glEnd();

	glPopMatrix();
}

/*
=================
GL_DrawCAR_Shadow
=================
*/
void GL_DrawCAR_Shadow (entity_t *e, int carnum)
{
	extern	cvar_t	gl_shadow;
	extern	int		ring_texture;
	extern	int		shadow_texture;

	float			xgrad, ygrad;

	glEnable (GL_BLEND);
	glDepthMask	(false);
	glColor4ub (cl.carcolor[carnum][0], cl.carcolor[carnum][1], cl.carcolor[carnum][2], 255);

	if ((gl_shadow.ivalue == 1) ||
		(gl_shadow.ivalue == 2 && e != &cl_entities[cl.viewentity]))
	{
		xgrad = 8 * (e->ground_normal[0] / e->ground_normal[2]);
		ygrad = 8 * (e->ground_normal[1] / e->ground_normal[2]);

		glBindTexture (GL_TEXTURE_2D, shadow_texture);

		glBegin (GL_QUADS);
			glTexCoord2f (0, 0); glVertex3f (e->ground_pos[0] - 8, e->ground_pos[1] - 8, e->ground_pos[2] + xgrad + ygrad);
			glTexCoord2f (0, 1); glVertex3f (e->ground_pos[0] - 8, e->ground_pos[1] + 8, e->ground_pos[2] + xgrad - ygrad);
			glTexCoord2f (1, 1); glVertex3f (e->ground_pos[0] + 8, e->ground_pos[1] + 8, e->ground_pos[2] - xgrad - ygrad);
			glTexCoord2f (1, 0); glVertex3f (e->ground_pos[0] + 8, e->ground_pos[1] - 8, e->ground_pos[2] - xgrad + ygrad);
		glEnd ();
	}
	else
	{
		xgrad = 16 * (e->ground_normal[0] / e->ground_normal[2]);
		ygrad = 16 * (e->ground_normal[1] / e->ground_normal[2]);
		
		glBindTexture (GL_TEXTURE_2D, ring_texture);

		glBegin (GL_QUADS);
			glTexCoord2f (0, 0); glVertex3f (e->ground_pos[0] - 16, e->ground_pos[1] - 16, e->ground_pos[2] + xgrad + ygrad);
			glTexCoord2f (0, 1); glVertex3f (e->ground_pos[0] - 16, e->ground_pos[1] + 16, e->ground_pos[2] + xgrad - ygrad);
			glTexCoord2f (1, 1); glVertex3f (e->ground_pos[0] + 16, e->ground_pos[1] + 16, e->ground_pos[2] - xgrad - ygrad);
			glTexCoord2f (1, 0); glVertex3f (e->ground_pos[0] + 16, e->ground_pos[1] - 16, e->ground_pos[2] - xgrad + ygrad);
		glEnd ();
	}
	glDepthMask	(true);
	glDisable (GL_BLEND);
}

/*
===========
GL_SetupCAR
===========
*/
void GL_SetupCAR (entity_t *e)
{
	extern	cvar_t	gl_shadow;
	vec3_t			mins, maxs;
	int				i;

	VectorAdd (e->origin, e->model->mins, mins);
	VectorAdd (e->origin, e->model->maxs, maxs);

	if (R_CullBox (mins, maxs))
		return;

	R_LightPoint (e->origin);

	i = e - cl_entities;

	if (!carskins[i-1] || cl.skinmodel[i-1] != e->model || !VectorCompare (cl.skincarcolor[i-1], cl.carcolor[i-1]))
		ColorSkin (i-1);

	glPushMatrix ();

    glTranslatef (e->origin[0], e->origin[1], e->origin[2]);

    glRotatef ( e->angles[1], 0, 0, 1);
    glRotatef (-e->angles[0], 0, 1, 0);
    glRotatef ( e->angles[2], 1, 0, 0);

	glColor4ub (lightcolor[0], lightcolor[1], lightcolor[2], 255);

	glEnableClientState (GL_VERTEX_ARRAY);

	glBindTexture (GL_TEXTURE_2D, carskins[i-1]);

	GL_DrawBody  (&e->model->carheader->Body);

	if (e->inair)
		glTranslatef (0, 0, -2);

	GL_DrawWheel (&e->model->carheader->Wheel[0]);
	GL_DrawWheel (&e->model->carheader->Wheel[1]);
	GL_DrawWheel (&e->model->carheader->Wheel[2]);
	GL_DrawWheel (&e->model->carheader->Wheel[3]);

	glDisableClientState (GL_VERTEX_ARRAY);

	glPopMatrix	();

	if (gl_shadow.ivalue)
  		GL_DrawCAR_Shadow (e, i-1);

	glColor4ub (255,255,255,255);
}

/*
===============
R_DrawMenuModel
===============
*/
void R_DrawMenuModel (int skin)
{
	extern	cvar_t	car;
	static	float	scale = 0;
	static	float	quad_rotate;
	static	model_t	*currentmodel;
	static	model_t	*oldmodel;
	model_t			*model;

	quad_rotate -= host_frametime*100;

	if (quad_rotate < 0)
		quad_rotate += 360;

	model = Mod_ForName (va("models/%s.car", car.string), true);

	if (currentmodel != model)
		scale = 0;

	glTranslatef (0,-7,-40);

	glRotatef (-75, 1,0,0);

	glRotatef (quad_rotate, 0,0,1);

	if (oldmodel != model)
	{
		oldmodel		= 0;
		currentmodel	= model;
		glScalef (scale/255, scale/255, scale/255);

		scale += host_frametime*200;

		if (scale > 255)
		{
			scale = 0;
			oldmodel = model;
		}
	}

	glEnableClientState (GL_VERTEX_ARRAY);

	glBindTexture (GL_TEXTURE_2D, skin);

	GL_DrawBody  (&model->carheader->Body);
	GL_DrawWheel (&model->carheader->Wheel[0]);
	GL_DrawWheel (&model->carheader->Wheel[1]);
	GL_DrawWheel (&model->carheader->Wheel[2]);
	GL_DrawWheel (&model->carheader->Wheel[3]);

	glDisableClientState (GL_VERTEX_ARRAY);
}
