/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"
#include	"../vid.h"
#include	"../crc.h"

typedef struct
{
	int				width;
	int				height;
	int				texnum;
	unsigned short	crc;
	char			identifier[64];
	qboolean		mipmap;
} gltexture_t;

#define	MAX_GLTEXTURES	256

gltexture_t	gltextures[MAX_GLTEXTURES];

cvar_t	gl_max_size = {"gl_max_size", "1024", true};

int		numgltextures;
int		image_width;
int		image_height;
byte	image_bpp;

int		sky_texture[5];
int		sign_texture[2];
int		water_texture[32];
int		countdown_texture[4];

int		icon_texture;
int		logo_texture;
int		ring_texture;
int		smoke_texture;
int		shadow_texture;
int		conback_texture;
int		conchars_texture;
int		speedometer_texture;

int		LoadMRTImage (char* filename, qboolean mipmap);

byte		*LoadTGA(FILE *fin);

void	GL_ClearSkids (void);
void	GL_ClearSmoke (void);

/*
=================
Draw_InitTextures
=================
*/
void Draw_InitTextures (void)
{
	icon_texture			= LoadMRTImage ("gfx/icon",			false);
	logo_texture			= LoadMRTImage ("gfx/logo",			false);
	ring_texture			= LoadMRTImage ("gfx/ring",			true);
	smoke_texture			= LoadMRTImage ("gfx/smoke",		true);
	shadow_texture			= LoadMRTImage ("gfx/shadow",		true);

	conback_texture			= LoadMRTImage ("gfx/conback",		false);
	conchars_texture		= LoadMRTImage ("gfx/conchars",		false);
	speedometer_texture		= LoadMRTImage ("gfx/speedometer",	false);

	sky_texture[0]			= LoadMRTImage ("gfx/skyft",		true);
	sky_texture[1]			= LoadMRTImage ("gfx/skybk",		true);
	sky_texture[2]			= LoadMRTImage ("gfx/skyrt",		true);
	sky_texture[3]			= LoadMRTImage ("gfx/skylf",		true);
	sky_texture[4]			= LoadMRTImage ("gfx/skyup",		true);

	sign_texture[0]			= LoadMRTImage ("gfx/sign0",		true);
	sign_texture[1]			= LoadMRTImage ("gfx/sign1",		true);

	water_texture[0]		= LoadMRTImage ("textures/water00",	true);
	water_texture[1]		= LoadMRTImage ("textures/water01",	true);
	water_texture[2]		= LoadMRTImage ("textures/water02",	true);
	water_texture[3]		= LoadMRTImage ("textures/water03",	true);
	water_texture[4]		= LoadMRTImage ("textures/water04",	true);
	water_texture[5]		= LoadMRTImage ("textures/water05",	true);
	water_texture[6]		= LoadMRTImage ("textures/water06",	true);
	water_texture[7]		= LoadMRTImage ("textures/water07",	true);
	water_texture[8]		= LoadMRTImage ("textures/water08",	true);
	water_texture[9]		= LoadMRTImage ("textures/water09",	true);
	water_texture[10]		= LoadMRTImage ("textures/water10",	true);
	water_texture[11]		= LoadMRTImage ("textures/water11",	true);
	water_texture[12]		= LoadMRTImage ("textures/water12",	true);
	water_texture[13]		= LoadMRTImage ("textures/water13",	true);
	water_texture[14]		= LoadMRTImage ("textures/water14",	true);
	water_texture[15]		= LoadMRTImage ("textures/water15",	true);
	water_texture[16]		= LoadMRTImage ("textures/water16",	true);
	water_texture[17]		= LoadMRTImage ("textures/water17",	true);
	water_texture[18]		= LoadMRTImage ("textures/water18",	true);
	water_texture[19]		= LoadMRTImage ("textures/water19",	true);
	water_texture[20]		= LoadMRTImage ("textures/water20",	true);
	water_texture[21]		= LoadMRTImage ("textures/water21",	true);
	water_texture[22]		= LoadMRTImage ("textures/water22",	true);
	water_texture[23]		= LoadMRTImage ("textures/water23",	true);
	water_texture[24]		= LoadMRTImage ("textures/water24",	true);
	water_texture[25]		= LoadMRTImage ("textures/water25",	true);
	water_texture[26]		= LoadMRTImage ("textures/water26",	true);
	water_texture[27]		= LoadMRTImage ("textures/water27",	true);
	water_texture[28]		= LoadMRTImage ("textures/water28",	true);
	water_texture[29]		= LoadMRTImage ("textures/water29",	true);
	water_texture[30]		= LoadMRTImage ("textures/water30",	true);
	water_texture[31]		= LoadMRTImage ("textures/water31",	true);

	countdown_texture[0]	= LoadMRTImage ("gfx/startgo",		false);
	countdown_texture[1]	= LoadMRTImage ("gfx/start1",		false);
	countdown_texture[2]	= LoadMRTImage ("gfx/start2",		false);
	countdown_texture[3]	= LoadMRTImage ("gfx/start3",		false);
}

float	char_array_tc[1024][8];
float	char_array_vt[1024][8];
byte	char_array_cl[1024][16];
int		char_arrays;

/*
=========
Draw_Init
=========
*/
void Draw_Init (void)
{
	int		maxsize;

	Cvar_RegisterVariable (&gl_max_size);

	Cmd_AddCommand ("clearskids", GL_ClearSkids);
	Cmd_AddCommand ("clearsmoke", GL_ClearSmoke);

	glGetIntegerv (GL_MAX_TEXTURE_SIZE, &maxsize);

	if (maxsize < gl_max_size.ivalue)
		Cvar_SetValue ("gl_max_size", maxsize);

	Draw_InitTextures ();

	char_arrays = 0;

	MR_memset (char_array_tc, 0, sizeof(char_array_tc));
	MR_memset (char_array_vt, 0, sizeof(char_array_vt));
	MR_memset (char_array_cl, 0, sizeof(char_array_cl));
}

/*
=============
Draw_Shutdown
=============
*/
extern	int	lightmap_textures[64];

void Draw_Shutdown (void)
{
	int		i;

	for (i = 0; i<numgltextures; ++i)
	{
		glDeleteTextures (1, & (gltextures[i].texnum));
		MR_memset(&gltextures[i], 0, sizeof(gltextures[i]));
	}

	numgltextures = 0;

	glDeleteTextures (64, lightmap_textures);
}

/*
===============
Draw_CharArrays
===============
*/
#ifdef _WIN32
extern void	(APIENTRY *MRglLockArraysEXT)		(GLint first, GLint count);
extern void	(APIENTRY *MRglUnlockArraysEXT)	(void);
#else
extern void	*MRglLockArraysEXT		(GLint first, GLint count);
extern void	*MRglUnlockArraysEXT	(void);
#endif

void Draw_CharArrays (void)
{
	glBindTexture (GL_TEXTURE_2D, conchars_texture);

	glTexEnvi	(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	glEnable	(GL_BLEND);

	glTexCoordPointer (2, GL_FLOAT,			8,	&char_array_tc[0][0]);
	glVertexPointer	  (2, GL_FLOAT,			8,	&char_array_vt[0][0]);
	glColorPointer	  (4, GL_UNSIGNED_BYTE,	4,	&char_array_cl[0][0]);

	glEnableClientState (GL_TEXTURE_COORD_ARRAY);
	glEnableClientState (GL_VERTEX_ARRAY);
	glEnableClientState (GL_COLOR_ARRAY);

	if(MRglLockArraysEXT)
		MRglLockArraysEXT (0, char_arrays * 4);

	glDrawArrays (GL_QUADS, 0, char_arrays * 4);

	if (MRglUnlockArraysEXT)
		MRglUnlockArraysEXT ();

	glDisableClientState (GL_TEXTURE_COORD_ARRAY);
	glDisableClientState (GL_VERTEX_ARRAY);
	glDisableClientState (GL_COLOR_ARRAY);

	glDisable	(GL_BLEND);
	glTexEnvi	(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);	
	glColor4ub	(255,255,255,255);

	MR_memset (char_array_tc, 0, sizeof(char_array_tc));
	MR_memset (char_array_vt, 0, sizeof(char_array_vt));
	MR_memset (char_array_cl, 0, sizeof(char_array_cl));
	char_arrays = 0;
}

/*
==============
Draw_Character
==============
*/
void Draw_Character (int x, int y, int num, byte scale, byte red, byte green, byte blue, byte alpha)
{
	float	row, col;

	if (num == 32 || y <= -8)
		return;		// space

	num &= 255;

	row = ((int)num >> 4) * 0.0625;
	col = ((int)num & 15) * 0.0625;

	char_array_tc[char_arrays][0] = char_array_tc[char_arrays][6] = col;
	char_array_tc[char_arrays][1] = char_array_tc[char_arrays][3] = row;
	char_array_tc[char_arrays][2] = char_array_tc[char_arrays][4] = col + 0.0625;
	char_array_tc[char_arrays][5] = char_array_tc[char_arrays][7] = row + 0.0625;

	char_array_vt[char_arrays][0] = char_array_vt[char_arrays][6] = x - scale;
	char_array_vt[char_arrays][1] = char_array_vt[char_arrays][3] = y - scale;
	char_array_vt[char_arrays][2] = char_array_vt[char_arrays][4] = x + scale + 16;
	char_array_vt[char_arrays][5] = char_array_vt[char_arrays][7] = y + scale + 16;

	char_array_cl[char_arrays][0]  = char_array_cl[char_arrays][4]  =
	char_array_cl[char_arrays][8]  = char_array_cl[char_arrays][12] = red;
	char_array_cl[char_arrays][1]  = char_array_cl[char_arrays][5]  =
	char_array_cl[char_arrays][9]  = char_array_cl[char_arrays][13] = green;
	char_array_cl[char_arrays][2]  = char_array_cl[char_arrays][6]  =
	char_array_cl[char_arrays][10] = char_array_cl[char_arrays][14] = blue;
	char_array_cl[char_arrays][3]  = char_array_cl[char_arrays][7]  =
	char_array_cl[char_arrays][11] = char_array_cl[char_arrays][15] = alpha;

	char_arrays++;

	if (char_arrays == 1024)
		Draw_CharArrays ();
}

/*
===========
Draw_String
===========
*/
void Draw_String (int x, int y, char *str, byte scale, byte red, byte green, byte blue, byte alpha)
{
	while (*str)
	{
		Draw_Character (x, y, *str, scale, red, green, blue, alpha);
		str++;
		x += 12;
	}
}

/*
======================
Draw_ConsoleBackground
======================
*/
void Draw_ConsoleBackground (void)
{
	int				change1 = (int)(host_frametime*100);
	int				change2 = MR_sin (realtime * 6.28318530717958647692f * 0.5) * 20;
	char			version[30];

	static qboolean	r_p = false;
	static qboolean	r_n = false;
	static qboolean	g_p = false;
	static qboolean	g_n = false;
	static qboolean	b_p = false;
	static qboolean	b_n = true;
	static byte		r = 255;
	static byte		g = 0;
	static byte		b = 0;

	if		(r_p && r == 255)	{r_p = false;	b_n = true;}
	else if (r_n && r == 0)		{r_n = false;	b_p = true;}
	else if (g_p && g == 255)	{g_p = false;	r_n = true;}
	else if (g_n && g == 0)		{g_n = false;	r_p = true;}
	else if (b_p && b == 255)	{b_p = false;	g_n = true;}
	else if (b_n && b == 0)		{b_n = false;	g_p = true;}

	if		(r_p)	r += change1;
	else if (r_n)	r -= change1;
	else if	(g_p)	g += change1;
	else if (g_n)	g -= change1;
	else if	(b_p)	b += change1;
	else if (b_n)	b -= change1;

	glBindTexture (GL_TEXTURE_2D, conback_texture);

	glColor4ub (r,g,b,255);

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	glBegin (GL_QUADS);
		glTexCoord2f (0, 0);	glVertex2f (0,			-20				- change2);
		glTexCoord2f (1, 0);	glVertex2f (vid.width,	-20				+ change2);
		glTexCoord2f (1, 1);	glVertex2f (vid.width,	vid.height + 20 + change2);
		glTexCoord2f (0, 1);	glVertex2f (0,			vid.height + 20 - change2);
	glEnd ();

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glColor4ub (255,255,255,255);

#ifdef _WIN32
	_snprintf (version, sizeof(version), "MiniRacer %.2f (win32)", (float) VERSION);
#else
	_snprintf (version, sizeof(version), "MiniRacer %.2f (linux)", (float) VERSION);
#endif

	//Draw_String (vid.width - 170, vid.height - 16, version, 0, 255, 255, 255, 255);
	Draw_String (vid.width - 280, vid.height - 16, version, 0, 255, 255, 255, 255);
}

/*
================
Draw_BeginDisc

Draws the little blue disc in the corner of the screen.
Call before beginning any disc IO.
================
*/
void Draw_BeginDisc (void)
{
	//do not draw icon before host is initialized
	if(host_initialized==true)
	{
		glBindTexture (GL_TEXTURE_2D, icon_texture);

		glBegin (GL_QUADS);
		    glTexCoord2f (0, 0);	glVertex2f (vid.width-64, 0);
		    glTexCoord2f (1, 0);	glVertex2f (vid.width, 0);
		    glTexCoord2f (1, 1);	glVertex2f (vid.width, 64);
		    glTexCoord2f (0, 1);	glVertex2f (vid.width-64, 64);
		glEnd ();
		
		glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

		glDrawBuffer(GL_FRONT);
		
		glDrawBuffer(GL_BACK);
	}
}

/*
================
Draw_EndDisc

Erases the disc icon.
Call after completing any disc IO
================
*/
void Draw_EndDisc (void)
{

}

//=============================================================================

/*
========
GL_Set2D
========
*/
void GL_Set2D (void)
{
	glViewport (0, 0, vid.width, vid.height);

	glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
	glOrtho  (0, vid.width, vid.height, 0, -99999, 99999);

	glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();

	glDisable (GL_DEPTH_TEST);
	glDisable (GL_CULL_FACE);
}

/*
=========
GL_Mipmap
=========
*/
void GL_Mipmap (byte *in, int width, int height)
{
	int		i, j;
	byte	*out;

	width <<=2;
	height >>= 1;
	out = in;
	for (i=0 ; i<height ; i++, in+=width)
	{
		for (j=0 ; j<width ; j+=8, out+=4, in+=8)
		{
			out[0] = (in[0] + in[4] + in[width+0] + in[width+4])>>2;
			out[1] = (in[1] + in[5] + in[width+1] + in[width+5])>>2;
			out[2] = (in[2] + in[6] + in[width+2] + in[width+6])>>2;
			out[3] = (in[3] + in[7] + in[width+3] + in[width+7])>>2;
		}
	}
}

/*
=========================
R_ResampleTextureLerpLine
=========================
*/
void R_ResampleTextureLerpLine (byte *in, byte *out, int inwidth, int outwidth)
{
	int		j, xi, oldx, f, fstep, endx;

	oldx	= 0;
	fstep	= (int) (inwidth * 65536.0f / outwidth);
	endx	= (inwidth - 1);

	for (j = 0,f = 0;j < outwidth;j++, f += fstep)
	{
		xi = (int) f >> 16;
		if (xi != oldx)
		{
			in += (xi - oldx) * 4;
			oldx = xi;
		}
		if (xi < endx)
		{
			int lerp = f & 0xFFFF;
			*out++ = (byte) ((((in[4] - in[0]) * lerp) >> 16) + in[0]);
			*out++ = (byte) ((((in[5] - in[1]) * lerp) >> 16) + in[1]);
			*out++ = (byte) ((((in[6] - in[2]) * lerp) >> 16) + in[2]);
			*out++ = (byte) ((((in[7] - in[3]) * lerp) >> 16) + in[3]);
		}
		else
		{
			*out++ = in[0];
			*out++ = in[1];
			*out++ = in[2];
			*out++ = in[3];
		}
	}
}

/*
=================
R_ResampleTexture
=================
*/
void R_ResampleTexture (void *indata, int inwidth, int inheight, void *outdata,  int outwidth, int outheight)
{
	int		i, j, yi, oldy, f, fstep, endy = (inheight-1);
	byte	*inrow, *out, *row1, *row2;

	out		= outdata;
	fstep	= (int)(inheight * 65536.0f / outheight);

	row1	= malloc (outwidth * 4);
	row2	= malloc (outwidth * 4);

	inrow	= indata;
	oldy	= 0;

	R_ResampleTextureLerpLine (inrow,				row1, inwidth, outwidth);
	R_ResampleTextureLerpLine (inrow + inwidth * 4, row2, inwidth, outwidth);

	for (i=0, f=0; i<outheight; i++, f+=fstep)
	{
		yi = f >> 16;
		if (yi < endy)
		{
			int lerp = f & 0xFFFF;

			if (yi != oldy)
			{
				inrow = (byte *)indata + inwidth * 4 * yi;
				if (yi == oldy+1)
				{
					MR_memcpy (row1, row2, outwidth * 4);
				}
				else
				{
					R_ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);
				}

				R_ResampleTextureLerpLine (inrow + inwidth*4, row2, inwidth, outwidth);
				oldy = yi;
			}

			j = outwidth - 4;

			while(j >= 0)
			{
				out[ 0] = (byte) ((((row2[ 0] - row1[ 0]) * lerp) >> 16) + row1[ 0]);
				out[ 1] = (byte) ((((row2[ 1] - row1[ 1]) * lerp) >> 16) + row1[ 1]);
				out[ 2] = (byte) ((((row2[ 2] - row1[ 2]) * lerp) >> 16) + row1[ 2]);
				out[ 3] = (byte) ((((row2[ 3] - row1[ 3]) * lerp) >> 16) + row1[ 3]);
				out[ 4] = (byte) ((((row2[ 4] - row1[ 4]) * lerp) >> 16) + row1[ 4]);
				out[ 5] = (byte) ((((row2[ 5] - row1[ 5]) * lerp) >> 16) + row1[ 5]);
				out[ 6] = (byte) ((((row2[ 6] - row1[ 6]) * lerp) >> 16) + row1[ 6]);
				out[ 7] = (byte) ((((row2[ 7] - row1[ 7]) * lerp) >> 16) + row1[ 7]);
				out[ 8] = (byte) ((((row2[ 8] - row1[ 8]) * lerp) >> 16) + row1[ 8]);
				out[ 9] = (byte) ((((row2[ 9] - row1[ 9]) * lerp) >> 16) + row1[ 9]);
				out[10] = (byte) ((((row2[10] - row1[10]) * lerp) >> 16) + row1[10]);
				out[11] = (byte) ((((row2[11] - row1[11]) * lerp) >> 16) + row1[11]);
				out[12] = (byte) ((((row2[12] - row1[12]) * lerp) >> 16) + row1[12]);
				out[13] = (byte) ((((row2[13] - row1[13]) * lerp) >> 16) + row1[13]);
				out[14] = (byte) ((((row2[14] - row1[14]) * lerp) >> 16) + row1[14]);
				out[15] = (byte) ((((row2[15] - row1[15]) * lerp) >> 16) + row1[15]);
				out		+= 16;
				row1	+= 16;
				row2	+= 16;
				j		-= 4;
			}
			if (j & 2)
			{
				out[ 0] = (byte) ((((row2[ 0] - row1[ 0]) * lerp) >> 16) + row1[ 0]);
				out[ 1] = (byte) ((((row2[ 1] - row1[ 1]) * lerp) >> 16) + row1[ 1]);
				out[ 2] = (byte) ((((row2[ 2] - row1[ 2]) * lerp) >> 16) + row1[ 2]);
				out[ 3] = (byte) ((((row2[ 3] - row1[ 3]) * lerp) >> 16) + row1[ 3]);
				out[ 4] = (byte) ((((row2[ 4] - row1[ 4]) * lerp) >> 16) + row1[ 4]);
				out[ 5] = (byte) ((((row2[ 5] - row1[ 5]) * lerp) >> 16) + row1[ 5]);
				out[ 6] = (byte) ((((row2[ 6] - row1[ 6]) * lerp) >> 16) + row1[ 6]);
				out[ 7] = (byte) ((((row2[ 7] - row1[ 7]) * lerp) >> 16) + row1[ 7]);
				out		+= 8;
				row1	+= 8;
				row2	+= 8;
			}
			if (j & 1)
			{
				out[ 0] = (byte) ((((row2[ 0] - row1[ 0]) * lerp) >> 16) + row1[ 0]);
				out[ 1] = (byte) ((((row2[ 1] - row1[ 1]) * lerp) >> 16) + row1[ 1]);
				out[ 2] = (byte) ((((row2[ 2] - row1[ 2]) * lerp) >> 16) + row1[ 2]);
				out[ 3] = (byte) ((((row2[ 3] - row1[ 3]) * lerp) >> 16) + row1[ 3]);
				out		+= 4;
				row1	+= 4;
				row2	+= 4;
			}
			row1 -= outwidth * 4;
			row2 -= outwidth * 4;
		}
		else
		{
			if (yi != oldy)
			{
				inrow = (byte *)indata + inwidth * 4 * yi;
				
				if (yi == oldy+1)
				{
					MR_memcpy (row1, row2, outwidth * 4);
				}
				else
				{
					R_ResampleTextureLerpLine (inrow, row1, inwidth, outwidth);
				}
				oldy = yi;
			}
			MR_memcpy (out, row1, outwidth * 4);
		}
	}

	free (row1);
	free (row2);
}

/*
===============
GL_UploadMipmap
===============
*/
void GL_UploadMipmap (unsigned *data, int width, int height, qboolean alpha, qboolean clamp)
{
	extern	cvar_t		gl_picmip;
	static	unsigned	scaled[1024 * 1024 * 4];
	int					scaled_width, scaled_height, type, miplevel;

	type		= alpha ? 4 : 3;
	miplevel	= 0;

	for (scaled_width  = 1; scaled_width  < width;  scaled_width  <<= 1);
	for (scaled_height = 1; scaled_height < height; scaled_height <<= 1);

	scaled_width  >>= (int)gl_picmip.ivalue;
	scaled_height >>= (int)gl_picmip.ivalue;

	if (scaled_width  > gl_max_size.ivalue)
		scaled_width  = gl_max_size.ivalue;
	if (scaled_height > gl_max_size.ivalue)
		scaled_height = gl_max_size.ivalue;

	if (scaled_width  > 1024)
		scaled_width  = 1024;
	if (scaled_height > 1024)
		scaled_height = 1024;

	if (scaled_width != width || scaled_height != height)
		R_ResampleTexture (data, width, height, scaled, scaled_width, scaled_height);
	else
		MR_memcpy (scaled, data, width * height * 4);

	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	if (clamp)
	{
		glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	}
	else
	{
		glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	}

	glTexImage2D (GL_TEXTURE_2D, 0, type, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);

	while (scaled_width > 1 || scaled_height > 1)
	{
		GL_Mipmap ((byte *)scaled, scaled_width, scaled_height);

		scaled_width  >>= 1;
		scaled_height >>= 1;

		if (scaled_width  < 1)
			scaled_width  = 1;
		if (scaled_height < 1)
			scaled_height = 1;

		miplevel++;

		if (clamp)
		{
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		}
		else
		{
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
			glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		}

		glTexImage2D (GL_TEXTURE_2D, miplevel, type, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);
	}
}

/*
=========
GL_Upload
=========
*/
void GL_Upload (unsigned *data, int width, int height, qboolean alpha)
{
	extern	cvar_t		gl_picmip;
	static	unsigned	scaled[1024 * 1024 * 4];
	int					scaled_width, scaled_height, type;

	type = alpha ? 4 : 3;

	for (scaled_width  = 1; scaled_width  < width;  scaled_width  <<= 1);
	for (scaled_height = 1; scaled_height < height; scaled_height <<= 1);

	scaled_width  >>= (int)gl_picmip.ivalue;
	scaled_height >>= (int)gl_picmip.ivalue;

	if (scaled_width  > gl_max_size.ivalue)
		scaled_width  = gl_max_size.ivalue;
	if (scaled_height > gl_max_size.ivalue)
		scaled_height = gl_max_size.ivalue;

	if (scaled_width  > 1024)
		scaled_width  = 1024;
	if (scaled_height > 1024)
		scaled_height = 1024;

	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameterf	(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

	if (scaled_width != width || scaled_height != height)
	{
		R_ResampleTexture (data, width, height, scaled, scaled_width, scaled_height);

		glTexImage2D (GL_TEXTURE_2D, 0, type, scaled_width, scaled_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, scaled);

		return;
	}
	
	glTexImage2D (GL_TEXTURE_2D, 0, type, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
}

/*
==============
GL_LoadTexture
==============
*/
int GL_LoadTexture (char *identifier, int width, int height, byte *data, qboolean mipmap, qboolean alpha)
{
	gltexture_t		*glt;
	int				i;
	unsigned short	crc;
	qboolean		clamp = false;

	if (!identifier[0])
		Host_Error("GL_LoadTexture: no identifier\n");

	crc = CRC_Block (data, width*height*4);

	for (i=0, glt=gltextures; i<numgltextures; i++, glt++)
	{
		if (!MR_strcmp (identifier, glt->identifier))
		{
			if (crc == glt->crc && width == glt->width && height == glt->height)
				return glt->texnum;
		}
	}

	glt = &gltextures[numgltextures];
	numgltextures++;

	MR_strcpy (glt->identifier, identifier);

	glGenTextures (1, &glt->texnum);

	glt->crc	= crc;
	glt->width	= width;
	glt->height	= height;
	glt->mipmap	= mipmap;

	if (!isDedicated)
	{
		if (glt->identifier[0] == 'c' &&
			glt->identifier[1] == 'u' &&
			glt->identifier[2] == 'r' &&
			glt->identifier[3] == 'v' &&
			glt->identifier[4] == 'e')
			clamp = true;

		glBindTexture (GL_TEXTURE_2D, glt->texnum);

		if (mipmap)	GL_UploadMipmap	((void *)data, width, height, alpha, clamp);
		else		GL_Upload		((void *)data, width, height, alpha);
	}

	return glt->texnum;
}

/*
===========
LoadMRTData

loads an image, currently only *.tga supported
===========
*/
byte *LoadMRTData (char* filename, qboolean complain)
{
	FILE	*f;
	char	name[128];

	_snprintf (name, sizeof(name), "%s.tga", filename);
	COM_FOpenFile(name, &f);
	if (f)
		return LoadTGA (f);

	if (complain)
		Con_Printf ("Couldn't load image (%s)\n", filename);
	return NULL;
}

/*
============
LoadMRTImage
============
*/
int LoadMRTImage (char* filename, qboolean mipmap)
{
	int			texnum;
	byte		*data;
	qboolean	alpha = false;

	data = LoadMRTData (filename, true);

	if(!data)
		return 0;

	if (image_bpp == 32)
		alpha = true;

	texnum = GL_LoadTexture (filename, image_width, image_height, data, mipmap, alpha);

	free (data);

	return texnum;
}

/*
=========================================================

TARGA LOADING

=========================================================
*/

typedef struct _TargaHeader {
	unsigned char 	id_length, colormap_type, image_type;
	unsigned short	colormap_index, colormap_length;
	unsigned char	colormap_size;
	unsigned short	x_origin, y_origin, width, height;
	unsigned char	pixel_size, attributes;
} TargaHeader;


//TargaHeader		targa_header;
//byte			*targa_rgba;

int fgetLittleShort (FILE *f)
{
	byte	b1, b2;

	b1 = fgetc(f);
	b2 = fgetc(f);

	return (short)(b1 + b2*256);
}

/*
=============
LoadTGA
=============
*/
byte *LoadTGA (FILE *fin)
{
	int				columns, rows, numPixels;
	byte			*pixbuf;
	int				row, column;
	static byte		*targa_rgba;
	TargaHeader		targa_header;

	targa_header.id_length = fgetc(fin);
	targa_header.colormap_type = fgetc(fin);
	targa_header.image_type = fgetc(fin);
	
	targa_header.colormap_index = fgetLittleShort(fin);
	targa_header.colormap_length = fgetLittleShort(fin);
	targa_header.colormap_size = fgetc(fin);
	targa_header.x_origin = fgetLittleShort(fin);
	targa_header.y_origin = fgetLittleShort(fin);
	targa_header.width = fgetLittleShort(fin);
	targa_header.height = fgetLittleShort(fin);
	targa_header.pixel_size = fgetc(fin);
	targa_header.attributes = fgetc(fin);

	//these are global variables
	image_width = targa_header.width;
	image_height = targa_header.height;
	image_bpp = targa_header.pixel_size;

	if (targa_header.image_type!=2 
		&& targa_header.image_type!=3 && targa_header.image_type!=10 && targa_header.image_type!=11)
		Sys_Error ("LoadTGA: Only type 2, 3, 10 and 11 targa images supported\n");

	if (targa_header.colormap_type !=0 
		|| (targa_header.pixel_size!=32 && targa_header.pixel_size!=24 && targa_header.pixel_size!=8))
		Sys_Error ("Texture_LoadTGA: Only 32, 24 bit RGB or 8bit greyscale images supported (no colormaps)\n");

	columns = targa_header.width;
	rows = targa_header.height;
	numPixels = columns * rows;

	targa_rgba = malloc (numPixels*4);
	
	if (targa_header.id_length != 0)
		fseek(fin, targa_header.id_length, SEEK_CUR);  // skip TARGA image comment
	
if (targa_header.attributes & 32)
//TOP TO BOTTOM
{
	if (targa_header.image_type==2) {  // Uncompressed, RGB images
		for(row=0; row<rows; row++) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; column++) {
				unsigned char red,green,blue,alphabyte;
				switch (targa_header.pixel_size) {
					case 24:
							
							blue = getc(fin);
							green = getc(fin);
							red = getc(fin);
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = 255;
							break;
					case 32:
							blue = getc(fin);
							green = getc(fin);
							red = getc(fin);
							alphabyte = getc(fin);
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = alphabyte;
							break;
				}
			}
		}    
	}
	else if (targa_header.image_type==3) {  // Uncompressed, greyscale images
		for(row=0; row<rows; row++) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; column++) {
				unsigned char red,green,blue;
							
				blue = green = red = getc(fin);
				*pixbuf++ = red;
				*pixbuf++ = green;
				*pixbuf++ = blue;
				*pixbuf++ = 255;	
			}
		}
	}
	else if (targa_header.image_type==10) {   // Runlength encoded RGB images
		unsigned char red,green,blue,alphabyte,packetHeader,packetSize,j;
		for(row=0; row<rows; row++) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; ) {
				packetHeader=getc(fin);
				packetSize = 1 + (packetHeader & 0x7f);
				if (packetHeader & 0x80) {        // run-length packet
					switch (targa_header.pixel_size) {
						case 24:
								blue = getc(fin);
								green = getc(fin);
								red = getc(fin);
								alphabyte = 255;
								break;
						case 32:
								blue = getc(fin);
								green = getc(fin);
								red = getc(fin);
								alphabyte = getc(fin);
								break;
					}
	
					for(j=0;j<packetSize;j++) {
						*pixbuf++=red;
						*pixbuf++=green;
						*pixbuf++=blue;
						*pixbuf++=alphabyte;
						column++;
						if (column==columns) { // run spans across rows
							column=0;
							if (row<rows-1)
								row++;
							else
								goto breakOut1032;
							pixbuf = targa_rgba + row*columns*4;
						}
					}
				}
				else {                            // non run-length packet
					for(j=0;j<packetSize;j++) {
						switch (targa_header.pixel_size) {
							case 24:
									blue = getc(fin);
									green = getc(fin);
									red = getc(fin);
									*pixbuf++ = red;
									*pixbuf++ = green;
									*pixbuf++ = blue;
									*pixbuf++ = 255;
									break;
							case 32:
									blue = getc(fin);
									green = getc(fin);
									red = getc(fin);
									alphabyte = getc(fin);
									*pixbuf++ = red;
									*pixbuf++ = green;
									*pixbuf++ = blue;
									*pixbuf++ = alphabyte;
									break;
						}
						column++;
						if (column==columns) { // pixel packet run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut1032;
							pixbuf = targa_rgba + row*columns*4;
						}						
					}
				}
			}
			breakOut1032:;
		}
	}
	else if (targa_header.image_type==11) {   // Runlength encoded greyscale images
		unsigned char red,green,blue,alphabyte,packetHeader,packetSize,j;
		for(row=0; row<rows; row++) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; ) {
				packetHeader=getc(fin);
				packetSize = 1 + (packetHeader & 0x7f);
				if (packetHeader & 0x80) {        // run-length packet
				
					blue = green = red = getc(fin);
					alphabyte = 255;
	
					for(j=0;j<packetSize;j++) {
						*pixbuf++=red;
						*pixbuf++=green;
						*pixbuf++=blue;
						*pixbuf++=alphabyte;
						column++;
						if (column==columns) { // run spans across rows
							column=0;
							if (row<rows-1)
								row++;
							else
								goto breakOut1132;
							pixbuf = targa_rgba + row*columns*4;
						}
					}
				}
				else {                            // non run-length packet
					for(j=0;j<packetSize;j++) {
					
						blue = green = red = getc(fin);
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = 255;
							
						column++;
						if (column==columns) { // pixel packet run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut1132;
							pixbuf = targa_rgba + row*columns*4;
						}						
					}
				}
			}
			breakOut1132:;
		}
	}
}
else
//BOTTOM TO TOP
{
	if (targa_header.image_type==2) {  // Uncompressed, RGB images
		for(row=rows-1; row>=0; row--) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; column++) {
				unsigned char red,green,blue,alphabyte;
				switch (targa_header.pixel_size) {
					case 24:
							
							blue = getc(fin);
							green = getc(fin);
							red = getc(fin);
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = 255;
							break;
					case 32:
							blue = getc(fin);
							green = getc(fin);
							red = getc(fin);
							alphabyte = getc(fin);
							*pixbuf++ = red;
							*pixbuf++ = green;
							*pixbuf++ = blue;
							*pixbuf++ = alphabyte;
							break;
				}
			}
		}    
	}
	else if (targa_header.image_type==3) {  // Uncompressed, greyscale images
		for(row=rows-1; row>=0; row--) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; column++) {
				unsigned char red,green,blue;
							
				blue = green = red = getc(fin);
				*pixbuf++ = red;
				*pixbuf++ = green;
				*pixbuf++ = blue;
				*pixbuf++ = 255;	
			}
		}
	}
	else if (targa_header.image_type==10) {   // Runlength encoded RGB images
		unsigned char red,green,blue,alphabyte,packetHeader,packetSize,j;
		for(row=rows-1; row>=0; row--) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; ) {
				packetHeader=getc(fin);
				packetSize = 1 + (packetHeader & 0x7f);
				if (packetHeader & 0x80) {        // run-length packet
					switch (targa_header.pixel_size) {
						case 24:
								blue = getc(fin);
								green = getc(fin);
								red = getc(fin);
								alphabyte = 255;
								break;
						case 32:
								blue = getc(fin);
								green = getc(fin);
								red = getc(fin);
								alphabyte = getc(fin);
								break;
					}
	
					for(j=0;j<packetSize;j++) {
						*pixbuf++=red;
						*pixbuf++=green;
						*pixbuf++=blue;
						*pixbuf++=alphabyte;
						column++;
						if (column==columns) { // run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut10;
							pixbuf = targa_rgba + row*columns*4;
						}
					}
				}
				else {                            // non run-length packet
					for(j=0;j<packetSize;j++) {
						switch (targa_header.pixel_size) {
							case 24:
									blue = getc(fin);
									green = getc(fin);
									red = getc(fin);
									*pixbuf++ = red;
									*pixbuf++ = green;
									*pixbuf++ = blue;
									*pixbuf++ = 255;
									break;
							case 32:
									blue = getc(fin);
									green = getc(fin);
									red = getc(fin);
									alphabyte = getc(fin);
									*pixbuf++ = red;
									*pixbuf++ = green;
									*pixbuf++ = blue;
									*pixbuf++ = alphabyte;
									break;
						}
						column++;
						if (column==columns) { // pixel packet run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut10;
							pixbuf = targa_rgba + row*columns*4;
						}						
					}
				}
			}
			breakOut10:;
		}
	}
	else if (targa_header.image_type==11) {   // Runlength encoded greyscale images
		unsigned char red,green,blue,alphabyte,packetHeader,packetSize,j;
		for(row=rows-1; row>=0; row--) {
			pixbuf = targa_rgba + row*columns*4;
			for(column=0; column<columns; ) {
				packetHeader=getc(fin);
				packetSize = 1 + (packetHeader & 0x7f);
				if (packetHeader & 0x80) {        // run-length packet
				
					blue = green = red = getc(fin);
					alphabyte = 255;
	
					for(j=0;j<packetSize;j++) {
						*pixbuf++=red;
						*pixbuf++=green;
						*pixbuf++=blue;
						*pixbuf++=alphabyte;
						column++;
						if (column==columns) { // run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut11;
							pixbuf = targa_rgba + row*columns*4;
						}
					}
				}
				else {                            // non run-length packet
					for(j=0;j<packetSize;j++) {
					
						blue = green = red = getc(fin);
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = 255;
							
						column++;
						if (column==columns) { // pixel packet run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut11;
							pixbuf = targa_rgba + row*columns*4;
						}						
					}
				}
			}
			breakOut11:;
		}
	}
}
	//fclose(fin);
	return targa_rgba;
}
