/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include	"../quakedef.h"
#include	"../vid.h"

mplane_t	frustum[4];

//
// view origin
//
vec3_t	vforward;
vec3_t	vright;
vec3_t	vup;

//
// screen size info
//
camera_t	camera;

mleaf_t		*r_viewleaf, *r_oldviewleaf;

void R_MarkLeaves (void);

cvar_t	numbots			= {"numbots",		"3", true};
cvar_t	numlaps			= {"numlaps",		"3", true};
cvar_t	draw_sky		= {"draw_sky",		"0", true};
cvar_t	gl_picmip		= {"gl_picmip",		"0", true};
cvar_t	gl_shadow		= {"gl_shadow",		"2", true};
cvar_t	draw_water		= {"draw_water",	"1", true};
cvar_t	_timeofday		= {"_timeofday",	"0", true};

byte* LoadMRTData (char* filename, qboolean complain);

int			carskins[8];

extern int	image_width;
extern int	image_height;

void ColorSkin (int carnum)
{
	int			i;
	char		filename[128];
	byte		*temp, *data, *in, *out;
	model_t		*model;
	entity_t	*entity;

	entity	= &cl_entities[1+carnum];
	model	= entity->model;

	if (!model)
		return;

	COM_StripExtension (model->name, filename);

	data	= LoadMRTData (filename, true);
	temp	= malloc (image_width*image_height*4);
	in		= data;
	out		= temp;

	for (i = 0;i < image_width*image_height;i++)
	{
		if (in[0] == in[1] && in[0] == in[2]) // any shade of grey
		{
			out[0]	= in[0]	* cl.carcolor[carnum][0] >> 8;
			out[1]	= in[1] * cl.carcolor[carnum][1] >> 8;
			out[2]	= in[2]	* cl.carcolor[carnum][2] >> 8;
		}
		else
		{
			out[0]	= in[0];
			out[1]	= in[1];
			out[2]	= in[2];
		}

		out[3]		= in[3];

		in	+= 4;
		out += 4;
	}

	carskins[carnum] = GL_LoadTexture(va("carskins%i",carnum), image_width, image_height, temp, true, false);

	cl.skincarcolor[carnum][0]	= cl.carcolor[carnum][0];
	cl.skincarcolor[carnum][1]	= cl.carcolor[carnum][1];
	cl.skincarcolor[carnum][2]	= cl.carcolor[carnum][2];
	cl.skinmodel[carnum]		= entity->model;

	free (data);
	free (temp);
}

/*
====================
R_DrawEntitiesOnList
====================
*/
void GL_SetupCAR (entity_t *e);

void R_DrawEntitiesOnList (void)
{
	int			i;
	entity_t	*entity;

	for (i=0; i<cl_numvisedicts_brush; i++)
	{
		entity = cl_visedicts_brush[i];

		if (entity->model->type == mod_brush)
			R_DrawBrushModel (entity);
	}

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	for (i=0; i<cl_numvisedicts_car; i++)
	{
		entity = cl_visedicts_car[i];

		if (entity->model->type == mod_car)
			GL_SetupCAR (entity);
	}

	glTexEnvi (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
}

void R_SetFrustum (void)
{
	int		i;

	// front side is visible

	VectorAdd (vforward, vright, frustum[0].normal);
	VectorSubtract (vforward, vright, frustum[1].normal);

	VectorAdd (vforward, vup, frustum[2].normal);
	VectorSubtract (vforward, vup, frustum[3].normal);

	for (i=0; i<4; i++)
	{
		frustum[i].type = PLANE_ANYZ;
		frustum[i].dist = DotProduct (camera.origin, frustum[i].normal);
		BoxOnPlaneSideClassify (&frustum[i]);
	}
}

/*
============
R_SetupFrame
============
*/
void R_SetupFrame (void)
{
	r_framecount++;

	AngleVectors (camera.angles, vforward, vright, vup);

// current viewleaf
	r_oldviewleaf	= r_viewleaf;
	r_viewleaf		= Mod_PointInLeaf (camera.origin, cl.worldmodel);
}

/*
=========
R_SetupGL
=========
*/
void R_SetupGL (void)
{
	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	glFrustum (-4, 4, -3, 3, 4, 4096);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glRotatef	(-90, 1, 0, 0);
	glRotatef	( 90, 0, 0, 1);
	glRotatef	(-camera.angles[2],  1, 0, 0);
	glRotatef	(-camera.angles[0],  0, 1, 0);
	glRotatef	(-camera.angles[1],  0, 0, 1);
	glTranslatef(-camera.origin[0],  -camera.origin[1],  -camera.origin[2]);

	glEnable	(GL_CULL_FACE);
	glEnable	(GL_DEPTH_TEST);
}

void GL_DrawSkids (void);
void GL_DrawSmoke (void);

/*
============
R_RenderView
============
*/
void R_RenderView (void)
{
	extern	msurface_t  *waterchain;
	extern	cvar_t		draw_water;
	
	R_SetupFrame ();

	R_SetFrustum ();

	R_SetupGL ();

	R_MarkLeaves ();

	R_DrawWorld ();

	R_DrawEntitiesOnList ();

	GL_DrawSkids ();

	GL_DrawSmoke ();

	if (waterchain && draw_water.ivalue)
		R_DrawWater (waterchain);

	waterchain	= NULL;
}

/*
================
R_SetupMenuModel
================
*/
void R_DrawMenuModel (int skin);

void R_SetupMenuModel (int skin)
{
	glClear (GL_DEPTH_BUFFER_BIT);

	glMatrixMode (GL_PROJECTION);
	glLoadIdentity ();

	glFrustum (-4, 4, -3, 3, 4, 4096);

	glMatrixMode (GL_MODELVIEW);
	glLoadIdentity ();

	glEnable	(GL_CULL_FACE);
	glEnable	(GL_DEPTH_TEST);

	R_DrawMenuModel (skin);
}
