/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#include	"../quakedef.h"
#include	"../vid.h"
#include	"../winminiracer.h"
#include	<commctrl.h>

#define		TEXTURE0_SGIS	0x835E
#define		TEXTURE1_SGIS	0x835F

#define		TEXTURE0_ARB	0x84C0
#define		TEXTURE1_ARB	0x84C1

//from bcc55/include/winuser.h
#ifndef CDS_FULLSCREEN
#define CDS_FULLSCREEN      0x00000004
#endif

DEVMODE		devmode;

viddef_t	vid;

MR_mode_t		mode;

HINSTANCE	hInstance;
HGLRC		baseRC;
HDC			baseDC;
HWND		window;
RECT		WindowRect;
WORD		sysgamma[3][256];
WORD		curgamma[3][256];

int			CenterX;
int			CenterY;

GLenum		TEXTURE0 = 0;
GLenum		TEXTURE1 = 0;

qboolean	vid_wassuspended	= false;

LONG	WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);

void	(APIENTRY *MRglMTexCoord2f)		(unsigned int, float, float);
void	(APIENTRY *MRglSelectTexture)		(unsigned int);
void	(APIENTRY *MRglLockArraysEXT)		(int first, int count);
void	(APIENTRY *MRglUnlockArraysEXT)	(void);

void	BuildGammaTable			(float gamma, float scale, float base, unsigned short *out);
void	AppActivate				(qboolean fActive);
void	VID_UpdateGamma			(qboolean force);
void	VID_RestoreSystemGamma	(void);
void	Draw_InitTextures		(void);
void	ClearAllStates			(void);
void	Draw_Shutdown			(void);
void	GL_Init					(void);

/*
============
CenterWindow
============
*/
void CenterWindow (void)
{
	CenterX = (GetSystemMetrics(SM_CXSCREEN) - vid.width)  * 0.5;
	CenterY = (GetSystemMetrics(SM_CYSCREEN) - vid.height) * 0.5;
	SetWindowPos (window, NULL, CenterX, CenterY, 0, 0,	SWP_NOSIZE | SWP_NOZORDER | SWP_SHOWWINDOW | SWP_DRAWFRAME);
}

/*
===========
VID_SetMode
===========
*/
void VID_SetMode (void)
{
	int						width, height, PixelFormat;
	DWORD					WindowStyle;
	WNDCLASS				wc;
	PIXELFORMATDESCRIPTOR	pfd;

	/* Register the frame class */
	MR_memset (&wc,		0, sizeof(wc));
	MR_memset (&pfd,	0, sizeof(pfd));
	MR_memset (&window,	0, sizeof(window));

	hInstance			= GetModuleHandle(NULL);

	wc.style			= CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
	wc.lpfnWndProc		= (WNDPROC)MainWndProc;
	wc.cbClsExtra		= 0;
	wc.cbWndExtra		= 0;
	wc.hInstance		= hInstance;
	wc.hIcon			= 0;
	wc.hCursor			= LoadCursor (NULL,IDC_ARROW);
	wc.hbrBackground	= NULL;
	wc.lpszMenuName		= NULL;
	wc.lpszClassName	= "MiniRacer";

    if (!RegisterClass (&wc))
		Sys_Error ("Couldn't register window class");

	WindowRect.left		= 0;
	WindowRect.top		= 0;
	WindowRect.right	= mode.width;
	WindowRect.bottom	= mode.height;

	vid.width			= mode.width;
	vid.height			= mode.height;

	// Set either the fullscreen or windowed mode

	if (mode.fullscreen)
	{
		MR_memset (&devmode, 0, sizeof(devmode));

		devmode.dmFields		= DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
		devmode.dmPelsWidth		= mode.width;
		devmode.dmPelsHeight	= mode.height;
		devmode.dmBitsPerPel	= mode.bpp;
		devmode.dmSize			= sizeof (devmode);

		if (ChangeDisplaySettings (&devmode, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL)
			Sys_Error ("Couldn't set fullscreen mode");

		WindowStyle		= WS_POPUP;
	}
	else
		WindowStyle		= WS_OVERLAPPED | WS_BORDER | WS_CAPTION | WS_SYSMENU;

	AdjustWindowRect (&WindowRect, WindowStyle, false);

	width	= WindowRect.right  - WindowRect.left;
	height	= WindowRect.bottom - WindowRect.top;

	// Create the window
	window = CreateWindow
		("MiniRacer",
		 "MiniRacer",
		 WindowStyle,
		 WindowRect.left,
		 WindowRect.top,
		 width,
		 height,
		 NULL,
		 NULL,
		 hInstance,
		 NULL);

	if (!window)
		Sys_Error ("Couldn't create window");

	ShowCursor (false);

	CenterWindow ();

	// Settings for the OpenGL window

	pfd.nSize			= sizeof(PIXELFORMATDESCRIPTOR);	// Size of this pixel format descriptor
	pfd.nVersion		= 1;								// The version of this data structure
	pfd.dwFlags			= PFD_DRAW_TO_WINDOW			    // Buffer supports drawing to window
						| PFD_SUPPORT_OPENGL				// Buffer supports OpenGL drawing
						| PFD_DOUBLEBUFFER;					// Supports double buffering
	pfd.iPixelType		= PFD_TYPE_RGBA;					// RGBA color format
	pfd.cColorBits		= mode.bpp;							// OpenGL color depth
	pfd.cDepthBits		= 16;								// Specifies the depth of the depth buffer
	pfd.iLayerType		= PFD_MAIN_PLANE;					// Main layer

	baseDC = GetDC (window);
	if (!baseDC)
		Sys_Error ("Couldn't create DC");

	PixelFormat = ChoosePixelFormat (baseDC, &pfd);
	if (!PixelFormat)
		Sys_Error ("Couldn't find pixelformat");

	if(!SetPixelFormat (baseDC, PixelFormat, &pfd))
		Sys_Error ("Couldn't set pixelformat");

	baseRC = wglCreateContext (baseDC);
	if (!baseRC)
		Sys_Error ("Couldn't create RC");

	if(!wglMakeCurrent (baseDC, baseRC))
		Sys_Error ("Couldn't activate DC and RC");

	ShowWindow			(window, SW_SHOW);
	SetForegroundWindow	(window);
	SetFocus			(window);
	UpdateWindow		(window);

	ClearAllStates ();
}

/*
===============
CheckExtensions
===============
*/
void CheckExtensions (void) 
{
	const char	*gl_extensions;

	MRglMTexCoord2f		= NULL;
	MRglSelectTexture		= NULL;
	MRglLockArraysEXT		= NULL;
	MRglUnlockArraysEXT	= NULL;

	gl_extensions = glGetString (GL_EXTENSIONS);

	Con_Printf ("GL_VENDOR: %s\n", glGetString (GL_VENDOR));
	Con_Printf ("GL_RENDERER: %s\n", glGetString (GL_RENDERER));

	// Check for Multitexture Extensions

	if (COM_CheckParm("-nomtex")) 
	{
		Con_Printf ("Multitexture Disabled\n");
	}
	else if (strstr (gl_extensions, "GL_ARB_multitexture")) 
	{
		Con_Printf ("Found GL_ARB_multitexture\n");
		MRglMTexCoord2f	= (void *) wglGetProcAddress("glMultiTexCoord2fARB");
		MRglSelectTexture	= (void *) wglGetProcAddress("glActiveTextureARB");
		TEXTURE0		= TEXTURE0_ARB;
		TEXTURE1		= TEXTURE1_ARB;
	} 
	else if (strstr (gl_extensions, "GL_SGIS_multitexture")) 
	{
		Con_Printf ("Found GL_SGIS_multitexture\n");
		MRglMTexCoord2f	= (void *) wglGetProcAddress ("glMTexCoord2fSGIS");
		MRglSelectTexture	= (void *) wglGetProcAddress ("glSelectTextureSGIS");
		TEXTURE0		= TEXTURE0_SGIS;
		TEXTURE1		= TEXTURE1_SGIS;
	}
	else
	{
		Con_Printf ("Multitexture Not Found\n");
	}

	// Check for Vertex Array Extensions

	if (COM_CheckParm("-nocva")) 
	{
		Con_Printf ("Compiled Vertex Arrays Disabled\n");
	}
	else if (strstr (gl_extensions, "GL_EXT_compiled_vertex_array"))
	{
		Con_Printf ("Found GL_EXT_compiled_vertex_array\n");
		MRglLockArraysEXT		= (void *) wglGetProcAddress ("glLockArraysEXT");
		MRglUnlockArraysEXT	= (void *) wglGetProcAddress ("glUnlockArraysEXT");
	}
	else
	{
		Con_Printf ("Compiled Vertex Arrays not found\n");
	}
}

/*
=======
GL_Init
======
*/
void GL_Init (void)
{
	CheckExtensions ();

	glEnable		(GL_TEXTURE_2D);
	glEnable		(GL_DITHER);
	glDisable		(GL_BLEND);
	glDisable		(GL_ALPHA_TEST);

	glHint			(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glCullFace		(GL_FRONT);
	glClearColor	(0,0,0,0);

	glBlendFunc		(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDepthFunc		(GL_LEQUAL);
	glDepthMask		(true);
	glDepthRange	(0,1);

	glLineWidth		(3);
	glFrustum		(-4, 4, -3, 3, 4, 4096);

	glTexEnvi		(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glShadeModel	(GL_SMOOTH);
	glPolygonMode	(GL_FRONT_AND_BACK, GL_FILL);

	glClear			(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

/*
===============
GL_EndRendering
===============
*/
void GL_EndRendering (void)
{
	glFinish ();
	SwapBuffers	(baseDC);
}

/*
============
VID_Shutdown
============
*/
void VID_Shutdown (void)
{
	if (baseRC)
	{
		if (!wglMakeCurrent(NULL,NULL))
			Con_Printf ("Could Not Release RC And DC");

		if (!wglDeleteContext(baseRC))
			Con_Printf ("Could Not Release RC");
	}

	if (baseDC)
	{
		if (!ReleaseDC (window, baseDC))
			Con_Printf ("Could Not Release DC");
	}

	if (window)
	{
		ShowWindow (window, SW_HIDE);

		if (!DestroyWindow (window))
			Con_Printf ("Could Not Destroy window");
	}

	if (!UnregisterClass ("MiniRacer", hInstance))
		Con_Printf ("Could Not Unregister Class");

	ChangeDisplaySettings (NULL, 0);
	ShowCursor (true);

	baseRC		= NULL;
	baseDC		= NULL;
	window		= NULL;
	hInstance	= NULL;

	VID_RestoreSystemGamma ();
}

/*
===============
scanqwertytokey
===============
*/
byte scanqwertytokey[128] = 
{ 
//  0		1		2		3		4		5		6			7		8		9		A		B		C		D		E			F
	0,		27,		'1',	'2',	'3',	'4',	'5',		'6',	'7',	'8',	'9',	'0',	'-',	'=',	K_BACKSPACE,9,		// 0
	'q',	'w',    'e',    'r',	't',    'y',	'u',		'i',	'o',	'p',	'[',	']',	13 ,    K_CTRL,	'a',		's',	// 1
	'd',	'f',    'g',	'h',	'j',    'k',    'l',		';',	'\'',	'`',	K_SHIFT,'\\',	'z',    'x',    'c',		'v',	// 2
	'b',	'n',    'm',    ',',	'.',    '/',    K_SHIFT,	'*',	K_ALT,	' ',	0  ,    K_F1,	K_F2,	K_F3,	K_F4,		K_F5,   // 3
	K_F6,	K_F7,	K_F8,	K_F9,	K_F10,	K_PAUSE,0,			K_HOME,	K_UP,	K_PGUP,	'-',	K_LEFT,	'5',	K_RIGHT,'+',		K_END,	// 4
	K_DOWN,	K_PGDN,	K_INS,	K_DEL,	0,		0,		0,			K_F11,	K_F12,	0,		0,		0,		0,		0,		0,			0,		// 5
	0,		0,		0,		0,		0  ,    0,		0,			0,		0,		0,		0,		0,		0  ,    0,		0,			0,		// 6
	0,		0,		0,		0,		0  ,    0,		0,			0,		0,		0,		0,		0,		0  ,    0,		0,			0		// 7
}; 

/*
===============
scandvoraktokey
===============
*/
byte scandvoraktokey[128] = 
{ 
//  0		1		2		3		4		5		6			7		8		9		A		B		C		D		E			F
	0,		27,		'1',	'2',	'3',	'4',	'5',		'6',	'7',	'8',	'9',	'0',	'[',	']',	K_BACKSPACE,9,		// 0
	'\'',	',',    '.',    'p',	'y',    'f',	'g',		'c',	'r',	'l',	'/',	'=',	13 ,    K_CTRL,	'a',		'o',	// 1
	'e',	'u',    'i',	'd',	'h',    't',    'n',		's',	'-',	'\\',	K_SHIFT,'\\',	';',    'q',    'j',		'k',	// 2
	'x',	'b',    'm',    'w',	'v',    'z',    K_SHIFT,	'*',	K_ALT,	' ',	0  ,    K_F1,	K_F2,	K_F3,	K_F4,		K_F5,   // 3
	K_F6,	K_F7,	K_F8,	K_F9,	K_F10,	K_PAUSE,0,			K_HOME,	K_UP,	K_PGUP,	'-',	K_LEFT,	'5',	K_RIGHT,'+',		K_END,	// 4
	K_DOWN,	K_PGDN,	K_INS,	K_DEL,	0,		0,		0,			K_F11,	K_F12,	0,		0,		0,		0,		0,		0,			0,		// 5
	0,		0,		0,		0,		0  ,    0,		0,			0,		0,		0,		0,		0,		0  ,    0,		0,			0,		// 6
	0,		0,		0,		0,		0  ,    0,		0,			0,		0,		0,		0,		0,		0  ,    0,		0,			0		// 7
};

/*
======
MapKey
======
*/
int MapKey (int key)
{
	key = (key>>16)&255;

	if (key > 127)
		return 0;

	if (COM_CheckParm("-dvorak"))
	{
		if (scandvoraktokey[key] == 0)
			Con_DPrintf("key 0x%02x has no translation\n", key);
		return scandvoraktokey[key];
	}
	else
	{
		if (scanqwertytokey[key] == 0)
			Con_DPrintf("key 0x%02x has no translation\n", key);
		return scanqwertytokey[key];
	}

	return 0;
}

/*
==============
ClearAllStates
==============
*/
void ClearAllStates (void)
{
	int		i;
	
	for (i=0 ; i<256 ; i++)
		Key_Event (i, false);

	Key_ClearStates ();
}

/*
===========
AppActivate
===========
*/
void AppActivate (qboolean fActive)
{
	ActiveApp = fActive;

	if (fActive)
	{
		if (mode.fullscreen)
		{
			if (vid_wassuspended) 
			{
				vid_wassuspended = false;
				ChangeDisplaySettings (&devmode, CDS_FULLSCREEN);
				ShowWindow (window, SW_SHOWNORMAL);
			}
			MoveWindow (window, 0, 0, devmode.dmPelsWidth, devmode.dmPelsHeight, false);
		}
		VID_UpdateGamma (true);
	}

	if (!fActive)
	{
		if (mode.fullscreen)
		{
			ChangeDisplaySettings (NULL, 0);
			vid_wassuspended = true;
		}
	}
}

/*
===========
MainWndProc
===========
*/
LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	LONG    lRet = 1;
	int		fActive;

	switch (uMsg)
	{
		case WM_SYSCHAR:
		case WM_SIZE:
		case WM_CREATE:
		case WM_MOVE:
		case WM_DESTROY:
		case WM_KILLFOCUS:
			break;

		case WM_KEYDOWN:
		case WM_SYSKEYDOWN:
			Key_Event (MapKey(lParam), true);
			break;

		case WM_KEYUP:
		case WM_SYSKEYUP:
			Key_Event (MapKey(lParam), false);
			break;

		case WM_LBUTTONDOWN:
			Key_Event (K_MOUSE1, true);
			break;

		case WM_LBUTTONUP:
			Key_Event (K_MOUSE1, false);
			break;

		case WM_RBUTTONDOWN:
			Key_Event (K_MOUSE2, true);
			break;

		case WM_RBUTTONUP:
			Key_Event (K_MOUSE2, false);
			break;

		case WM_MBUTTONDOWN:
			Key_Event (K_MOUSE3, true);
			break;

		case WM_MBUTTONUP:
			Key_Event (K_MOUSE3, false);
			break;

   	    case WM_CLOSE:
			if (MessageBox (window, "Are you sure you want to quit?", "Confirm Exit", MB_YESNO | MB_SETFOREGROUND | MB_ICONQUESTION) == IDYES)
				Sys_Quit ();
	        break;

		case WM_ACTIVATE:
			fActive = LOWORD(wParam);

			AppActivate (!(fActive == WA_INACTIVE));

			ClearAllStates ();

			break;

    	default:
            /* pass all unhandled messages to DefWindowProc */
            lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    /* return 1 if handled message, 0 if not */
    return lRet;
}

/*
============
VID_SetGamma
============
*/
int VID_SetGamma (float gamma, float scale, float base)
{
	HDC hdc	= GetDC (NULL);
	int i;

	BuildGammaTable (gamma, scale, base, &curgamma[0][0]);

	for (i = 0;i < 256;i++)
		curgamma[1][i] = curgamma[2][i] = curgamma[0][i];

	i = SetDeviceGammaRamp (hdc, &curgamma[0][0]);

	ReleaseDC (NULL, hdc);
	return i;
}

/*
==================
VID_GetSystemGamma
==================
*/
void VID_GetSystemGamma (void)
{
	HDC hdc = GetDC (NULL);

	GetDeviceGammaRamp (hdc, &sysgamma[0][0]);

	ReleaseDC (NULL, hdc);
}

/*
======================
VID_RestoreSystemGamma
======================
*/
void VID_RestoreSystemGamma (void)
{
	HDC hdc = GetDC (NULL);

	SetDeviceGammaRamp (hdc, &sysgamma[0][0]);

	ReleaseDC (NULL, hdc);
}

/*
========
VID_Init
========
*/
void VID_Init (void)
{
	int		i;

	VID_GetSystemGamma ();

	mode.width		= 640;
	mode.height		= 480;
	mode.bpp		= 16;
	mode.fullscreen	= true;

	i = COM_CheckParm ("-width");
	if (i)
		mode.width		= MR_atoi (com_argv[i+1]);

	i = COM_CheckParm ("-height");
	if (i)
		mode.height		= MR_atoi (com_argv[i+1]);

	i = COM_CheckParm ("-bpp");
	if (i)
		mode.bpp		= MR_atoi (com_argv[i+1]);

	i = COM_CheckParm ("-window");
	if (i)
		mode.fullscreen	= false;

	InitCommonControls();

	VID_SetMode ();

	GL_Init ();
}

