/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sv_user.c -- server code for moving users

#include	"../quakedef.h"

edict_t	*sv_player;

/*
===================
SV_ReadClientMove
===================
*/
void SV_ReadClientMove ()
{
	int	bits;
	
// read ping time
	host_client->ping_times[host_client->num_pings%NUM_PING_TIMES] = sv.time - MSG_ReadFloat ();
	host_client->num_pings++;
		
	bits = MSG_ReadByte ();
	host_client->edict->game.forward	= (bits >> 0)& 1;
	host_client->edict->game.back		= (bits >> 1)& 1;
	host_client->edict->game.left		= (bits >> 2)& 1;
	host_client->edict->game.right		= (bits >> 3)& 1;
	host_client->edict->game.brake		= (bits >> 4)& 1;
	host_client->edict->game.turbo		= (bits >> 5)& 1;

	if (bits & 64)	
		host_client->edict->game.impulse = MSG_ReadByte ();
}

/*
====================
SV_ReadClientMessage

Returns false if the client should be killed
====================
*/
qboolean SV_ReadClientMessage (void)
{
	int		ret;
	int		cmd;
	char		*s;
	
	do
	{
nextmsg:
		ret = NET_GetMessage (host_client->netconnection);
		if (ret == -1)
		{
			Sys_Printf ("SV_ReadClientMessage: NET_GetMessage failed\n");
			return false;
		}
		if (!ret)
			return true;
					
		MSG_BeginReading ();
		
		while (1)
		{
			if (!host_client->active)
				return false;	// a command caused an error

			if (msg_badread)
			{
				Sys_Printf ("SV_ReadClientMessage: badread\n");
				return false;
			}	
	
			cmd = MSG_ReadByte ();
			
			switch (cmd)
			{
			case -1:
				goto nextmsg;		// end of message
				
			default:
				Sys_Printf ("SV_ReadClientMessage: unknown command char\n");
				return false;
							
			case clc_nop:
//				Sys_Printf ("clc_nop\n");
				break;
				
			case clc_stringcmd:	
				s	= MSG_ReadString ();
				ret = 0;

				if (MR_strncasecmp(s, "status", 6) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "name", 4) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "say", 3) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "tell", 4) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "_carcolor", 9) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "_carmodel", 9) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "kill", 4) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "spawn", 5) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "begin", 5) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "prespawn", 8) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "kick", 4) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "ping", 4) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "give", 4) == 0)
					ret = 1;
				else if (MR_strncasecmp(s, "ban", 3) == 0)
					ret = 1;
				if (ret == 2)
					Cbuf_InsertText (s);
				else if (ret == 1)
					Cmd_ExecuteString (s, src_client);
				else
					Con_DPrintf("%s tried to %s\n", host_client->name, s);
				break;
				
			case clc_disconnect:
//				Sys_Printf ("SV_ReadClientMessage: client disconnected\n");
				return false;
			
			case clc_move:
				SV_ReadClientMove ();
				break;
			}
		}
	} while (ret == 1);
	
	return true;
}


/*
==================
SV_RunClients
==================
*/
void SV_RunClients (void)
{
	int				i;
	
	for (i=0, host_client = svs.clients ; i<svs.maxclients ; i++, host_client++)
	{
		if (!host_client->active)
			continue;
	
		sv_player = host_client->edict;

		if (!SV_ReadClientMessage ())
		{
			SV_DropClient (false);	// client misbehaved...
			continue;
		}
	}
}
